import matplotlib.pyplot as plt
import numpy as np
from skimage.exposure import rescale_intensity


def plot_vs_n_fluor(vs_nucleus, vs_membrane, fluor_nucleus, fluor_membrane):
    colormap_1 = [0.1254902, 0.6784314, 0.972549]  # bop blue
    colormap_2 = [0.972549, 0.6784314, 0.1254902]  # bop orange
    colormap_3 = [0, 1, 0]  # green
    colormap_4 = [1, 0, 1]  # magenta

    # Rescale the intensity
    vs_nucleus = rescale_intensity(vs_nucleus, out_range=(0, 1))
    vs_membrane = rescale_intensity(vs_membrane, out_range=(0, 1))
    # VS Nucleus RGB
    vs_nucleus_rgb = np.zeros((*vs_nucleus.shape[-2:], 3))
    vs_nucleus_rgb[:, :, 0] = vs_nucleus * colormap_1[0]
    vs_nucleus_rgb[:, :, 1] = vs_nucleus * colormap_1[1]
    vs_nucleus_rgb[:, :, 2] = vs_nucleus * colormap_1[2]
    # VS Membrane RGB
    vs_membrane_rgb = np.zeros((*vs_membrane.data.shape[-2:], 3))
    vs_membrane_rgb[:, :, 0] = vs_membrane * colormap_2[0]
    vs_membrane_rgb[:, :, 1] = vs_membrane * colormap_2[1]
    vs_membrane_rgb[:, :, 2] = vs_membrane * colormap_2[2]
    # Merge the two channels
    merged_vs = np.zeros((*vs_nucleus.shape[-2:], 3))
    merged_vs[:, :, 0] = vs_nucleus * colormap_1[0] + vs_membrane * colormap_2[0]
    merged_vs[:, :, 1] = vs_nucleus * colormap_1[1] + vs_membrane * colormap_2[1]
    merged_vs[:, :, 2] = vs_nucleus * colormap_1[2] + vs_membrane * colormap_2[2]

    # Rescale the intensity
    fluor_nucleus = rescale_intensity(fluor_nucleus, out_range=(0, 1))
    fluor_membrane = rescale_intensity(fluor_membrane, out_range=(0, 1))
    # fluor Nucleus RGB
    fluor_nucleus_rgb = np.zeros((*fluor_nucleus.shape[-2:], 3))
    fluor_nucleus_rgb[:, :, 0] = fluor_nucleus * colormap_3[0]
    fluor_nucleus_rgb[:, :, 1] = fluor_nucleus * colormap_3[1]
    fluor_nucleus_rgb[:, :, 2] = fluor_nucleus * colormap_3[2]
    # fluor Membrane RGB
    fluor_membrane_rgb = np.zeros((*fluor_membrane.shape[-2:], 3))
    fluor_membrane_rgb[:, :, 0] = fluor_membrane * colormap_4[0]
    fluor_membrane_rgb[:, :, 1] = fluor_membrane * colormap_4[1]
    fluor_membrane_rgb[:, :, 2] = fluor_membrane * colormap_4[2]
    # Merge the two channels
    merged_fluor = np.zeros((*fluor_nucleus.shape[-2:], 3))
    merged_fluor[:, :, 0] = (
        fluor_nucleus * colormap_3[0] + fluor_membrane * colormap_4[0]
    )
    merged_fluor[:, :, 1] = (
        fluor_nucleus * colormap_3[1] + fluor_membrane * colormap_4[1]
    )
    merged_fluor[:, :, 2] = (
        fluor_nucleus * colormap_3[2] + fluor_membrane * colormap_4[2]
    )

    # %%
    # Plot
    fig, ax = plt.subplots(2, 3, figsize=(15, 10))

    # Virtual staining plots
    ax[0, 0].imshow(vs_nucleus_rgb)
    ax[0, 0].set_title("VS Nuclei")
    ax[0, 1].imshow(vs_membrane_rgb)
    ax[0, 1].set_title("VS Membrane")
    ax[0, 2].imshow(merged_vs)
    ax[0, 2].set_title("VS Nuclei+Membrane")

    # Experimental fluorescence plots
    ax[1, 0].imshow(fluor_nucleus_rgb)
    ax[1, 0].set_title("Experimental Fluorescence Nuclei")
    ax[1, 1].imshow(fluor_membrane_rgb)
    ax[1, 1].set_title("Experimental Fluorescence Membrane")
    ax[1, 2].imshow(merged_fluor)
    ax[1, 2].set_title("Experimental Fluorescence Nuclei+Membrane")

    # turnoff axis
    for a in ax.flatten():
        a.axis("off")
    plt.margins(0, 0)
    plt.show()



""" Script to compute the correlation between PCA and UMAP features and computed features
* finds the computed features best representing the PCA and UMAP components
* outputs a heatmap of the correlation between PCA and UMAP features and computed features
"""

# %%
from pathlib import Path
import matplotlib.pyplot as plt
import seaborn as sns
from compute_pca_features import compute_features, compute_correlation_and_save_png

# %% for sensor features

features_path = Path(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/trainng_logs/SEC61/rev6_NTXent_sensorPhase_infection/2chan_160patch_94ckpt_rev6_2.zarr"
)
data_path = Path(
    "/hpc/projects/intracellular_dashboard/organelle_dynamics/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/registered_test.zarr"
)
tracks_path = Path(
    "/hpc/projects/intracellular_dashboard/organelle_dynamics/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/track_test.zarr"
)

source_channel = ["Phase3D", "RFP"]
seg_channel = ["Nuclei_prediction_labels"]
z_range = (28, 43)
fov_list = ["/A/3", "/B/3", "/B/4"]

features_sensor = compute_features(
    features_path,
    data_path,
    tracks_path,
    source_channel,
    seg_channel,
    z_range,
    fov_list,
)

features_sensor.to_csv(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/features_allset_sensor.csv",
    index=False,
)

# features_sensor = pd.read_csv("/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/features_allset_sensor.csv")

# take a subset without the 768 features
feature_columns = [f"feature_{i+1}" for i in range(768)]
features_subset_sensor = features_sensor.drop(columns=feature_columns)
correlation_sensor = compute_correlation_and_save_png(
    features_subset_sensor,
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/PC_vs_CF_2chan_pca_sensor_allset.svg",
)

# %% plot PCA vs set of computed features for sensor features

set_features = [
    "Fluor Radial Intensity Gradient",
    "Phase Interquartile Range",
    "Perimeter area ratio",
    "Fluor Interquartile Range",
    "Phase Entropy",
    "Fluor Zernike Moment Mean",
]

plt.figure(figsize=(10, 8))
sns.heatmap(
    correlation_sensor.loc[set_features, "PCA1":"PCA6"],
    annot=True,
    cmap="coolwarm",
    fmt=".2f",
    annot_kws={"size": 24},
    vmin=-1,
    vmax=1,
)
plt.xlabel("Computed Features", fontsize=24)
plt.ylabel("PCA Features", fontsize=24)
plt.xticks(fontsize=24)  # Increase x-axis tick labels
plt.yticks(fontsize=24)  # Increase y-axis tick labels

plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/PC_vs_CF_2chan_pca_allset_sensor_6features.svg"
)

# plot the PCA1 vs PCA2 map for sensor features

plt.figure(figsize=(10, 10))
sns.scatterplot(
    x="PCA1",
    y="PCA2",
    data=features_sensor,
)


#   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.   .-.-.
#  / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \ / / \ \
# '-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'-'   '-'


# %% for organelle features

features_path = Path(
    "/hpc/projects/intracellular_dashboard/organelle_dynamics/2024_11_07_A549_SEC61_ZIKV_DENV/4-phenotyping/predictions/Soorya/timeAware_2chan_ntxent_192patch_91ckpt_rev7_GT.zarr"
)
data_path = Path(
    "/hpc/projects/intracellular_dashboard/organelle_dynamics/2024_11_07_A549_SEC61_ZIKV_DENV/2-assemble/2024_11_07_A549_SEC61_ZIKV_DENV.zarr"
)
tracks_path = Path(
    "/hpc/projects/intracellular_dashboard/organelle_dynamics/2024_11_07_A549_SEC61_ZIKV_DENV/1-preprocess/label-free/4-track-gt/2024_11_07_A549_SEC61_ZIKV_DENV_2_cropped.zarr"
)

source_channel = ["Phase3D", "raw GFP EX488 EM525-45"]
seg_channel = ["nuclei_prediction_labels_labels"]
z_range = (16, 21)
normalizations = None
fov_list = ["/B/2/000000", "/B/3/000000", "/C/2/000000"]

features_organelle = compute_features(
    features_path,
    data_path,
    tracks_path,
    source_channel,
    seg_channel,
    z_range,
    fov_list,
)

# Save the features dataframe to a CSV file
features_organelle.to_csv(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/features_twoChan_organelle_multiwell.csv",
    index=False,
)

correlation_organelle = compute_correlation_and_save_png(
    features_organelle,
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/PC_vs_CF_2chan_pca_organelle_multiwell.svg",
)

# features_organelle = pd.read_csv("/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/features_twoChan_organelle_multiwell_refinedPCA.csv")

# %% plot PCA vs set of computed features for organelle features

plt.figure(figsize=(10, 8))
sns.heatmap(
    correlation_organelle.loc[set_features, "PCA1":"PCA6"],
    annot=True,
    cmap="coolwarm",
    fmt=".2f",
    annot_kws={"size": 24},
    vmin=-1,
    vmax=1,
)
plt.xlabel("Computed Features", fontsize=24)
plt.ylabel("PCA Features", fontsize=24)
plt.xticks(fontsize=24)  # Increase x-axis tick labels
plt.yticks(fontsize=24)  # Increase y-axis tick labels
plt.savefig(
    "/hpc/projects/comp.micro/infected_cell_imaging/Single_cell_phenotyping/ContrastiveLearning/Figure_panels/cell_division/PC_vs_CF_2chan_pca_setfeatures_organelle_6features.svg"
)

# %%



"""
Generate embeddings using a pre-trained ImageNet model and save them to a zarr store
using VisCy Trainer and EmbeddingWriter callback.
"""

import importlib
import logging
import os
from pathlib import Path
from typing import Dict, List, Literal, Optional

import click
import timm
import torch
import yaml
from lightning.pytorch import LightningModule

from viscy.data.triplet import TripletDataModule
from viscy.representation.embedding_writer import EmbeddingWriter
from viscy.trainer import VisCyTrainer

logger = logging.getLogger(__name__)


class ImageNetModule(LightningModule):
    def __init__(
        self,
        model_name: str = "convnext_tiny",
        channel_reduction_methods: Optional[
            Dict[str, Literal["middle_slice", "mean", "max"]]
        ] = None,
        channel_names: Optional[List[str]] = None,
    ):
        """Initialize the ImageNet module.

        Args:
            model_name: Name of the pre-trained ImageNet model to use
            channel_reduction_methods: Dict mapping channel names to reduction methods:
                - "middle_slice": Take the middle slice along the depth dimension
                - "mean": Average across the depth dimension
                - "max": Take the maximum value across the depth dimension
            channel_names: List of channel names corresponding to the input channels
        """
        super().__init__()
        self.channel_reduction_methods = channel_reduction_methods or {}
        self.channel_names = channel_names or []

        try:
            torch.set_float32_matmul_precision("high")
            self.model = timm.create_model(model_name, pretrained=True)
            self.model.eval()
        except ImportError:
            raise ImportError("Please install the timm library: " "pip install timm")

    def _reduce_5d_input(self, x: torch.Tensor) -> torch.Tensor:
        """Reduce 5D input (B, C, D, H, W) to 4D (B, C, H, W) using specified methods.

        Args:
            x: 5D input tensor

        Returns:
            4D tensor after applying reduction methods
        """
        if x.dim() != 5:
            return x

        B, C, D, H, W = x.shape
        result = torch.zeros((B, C, H, W), device=x.device)

        # Process all channels at once for each reduction method to minimize loops
        middle_slice_indices = []
        mean_indices = []
        max_indices = []

        # Group channels by reduction method
        for c in range(C):
            channel_name = (
                self.channel_names[c] if c < len(self.channel_names) else f"channel_{c}"
            )
            method = self.channel_reduction_methods.get(channel_name, "middle_slice")

            if method == "mean":
                mean_indices.append(c)
            elif method == "max":
                max_indices.append(c)
            else:  # Default to middle_slice for any unknown method
                middle_slice_indices.append(c)

        # Apply middle_slice reduction to all relevant channels at once
        if middle_slice_indices:
            indices = torch.tensor(middle_slice_indices, device=x.device)
            result[:, indices] = x[:, indices, D // 2]

        # Apply mean reduction to all relevant channels at once
        if mean_indices:
            indices = torch.tensor(mean_indices, device=x.device)
            result[:, indices] = x[:, indices].mean(dim=2)

        # Apply max reduction to all relevant channels at once
        if max_indices:
            indices = torch.tensor(max_indices, device=x.device)
            result[:, indices] = x[:, indices].max(dim=2)[0]

        return result

    def _convert_to_rgb(self, x: torch.Tensor) -> torch.Tensor:
        """Convert input tensor to 3-channel RGB format as needed.

        Args:
            x: Input tensor with 1, 2, or 3+ channels

        Returns:
            3-channel tensor suitable for ImageNet models
        """
        if x.shape[1] == 3:
            return x
        elif x.shape[1] == 1:
            # Convert to RGB by repeating the channel 3 times
            return x.repeat(1, 3, 1, 1)
        elif x.shape[1] == 2:
            # Normalize each channel independently to handle different scales
            B, _, H, W = x.shape
            x_3ch = torch.zeros((B, 3, H, W), device=x.device, dtype=x.dtype)

            # Normalize each channel to 0-1 range
            ch0 = x[:, 0:1]
            ch1 = x[:, 1:2]

            ch0_min = ch0.reshape(B, -1).min(dim=1, keepdim=True)[0].reshape(B, 1, 1, 1)
            ch0_max = ch0.reshape(B, -1).max(dim=1, keepdim=True)[0].reshape(B, 1, 1, 1)
            ch0_range = ch0_max - ch0_min + 1e-7  # Add epsilon for numerical stability
            ch0_norm = (ch0 - ch0_min) / ch0_range

            ch1_min = ch1.reshape(B, -1).min(dim=1, keepdim=True)[0].reshape(B, 1, 1, 1)
            ch1_max = ch1.reshape(B, -1).max(dim=1, keepdim=True)[0].reshape(B, 1, 1, 1)
            ch1_range = ch1_max - ch1_min + 1e-7  # Add epsilon for numerical stability
            ch1_norm = (ch1 - ch1_min) / ch1_range

            # Create blended RGB channels - map each normalized channel to different colors
            x_3ch[:, 0] = ch0_norm.squeeze(1)  # R channel from first input
            x_3ch[:, 1] = ch1_norm.squeeze(1)  # G channel from second input
            x_3ch[:, 2] = 0.5 * (
                ch0_norm.squeeze(1) + ch1_norm.squeeze(1)
            )  # B channel as blend

            return x_3ch
        else:
            # For more than 3 channels, use the first 3
            return x[:, :3]

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        """Extract features from the input images.

        Returns:
            Dictionary with features, properly shaped empty projections tensor, and index information
        """
        x = batch["anchor"]

        # Handle 5D input (B, C, D, H, W) using configured reduction methods
        if x.dim() == 5:
            x = self._reduce_5d_input(x)

        # Convert input to RGB format
        x = self._convert_to_rgb(x)

        # Get embeddings
        with torch.no_grad():
            features = self.model.forward_features(x)

            # Average pooling to get feature vector
            if features.dim() > 2:
                features = features.mean(dim=[2, 3])

        # Return features and empty projections with correct batch dimension
        return {
            "features": features,
            "projections": torch.zeros((features.shape[0], 0), device=features.device),
            "index": batch["index"],
        }


def load_config(config_file):
    """Load configuration from a YAML file."""
    with open(config_file, "r") as f:
        config = yaml.safe_load(f)
    return config


def load_normalization_from_config(norm_config):
    """Load a normalization transform from a configuration dictionary."""
    class_path = norm_config["class_path"]
    init_args = norm_config.get("init_args", {})

    # Split module and class name
    module_path, class_name = class_path.rsplit(".", 1)

    # Import the module
    module = importlib.import_module(module_path)

    # Get the class
    transform_class = getattr(module, class_name)

    # Instantiate the transform
    return transform_class(**init_args)


@click.command()
@click.option(
    "--config",
    "-c",
    type=click.Path(exists=True),
    required=True,
    help="Path to YAML configuration file",
)
@click.option(
    "--model",
    "-m",
    type=str,
    default="convnext_tiny",
    help="Name of the pre-trained ImageNet model to use",
)
def main(config, model):
    """Extract ImageNet embeddings and save to zarr format using VisCy Trainer."""
    # Configure logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    # Load config file
    cfg = load_config(config)
    logger.info(f"Loaded configuration from {config}")

    # Prepare datamodule parameters
    dm_params = {}

    # Add data and tracks paths from the paths section
    if "paths" not in cfg:
        raise ValueError("Configuration must contain a 'paths' section")

    if "data_path" not in cfg["paths"]:
        raise ValueError(
            "Data path is required in the configuration file (paths.data_path)"
        )
    dm_params["data_path"] = cfg["paths"]["data_path"]

    if "tracks_path" not in cfg["paths"]:
        raise ValueError(
            "Tracks path is required in the configuration file (paths.tracks_path)"
        )
    dm_params["tracks_path"] = cfg["paths"]["tracks_path"]

    # Add datamodule parameters
    if "datamodule" not in cfg:
        raise ValueError("Configuration must contain a 'datamodule' section")

    # Prepare normalizations
    if (
        "normalizations" not in cfg["datamodule"]
        or not cfg["datamodule"]["normalizations"]
    ):
        raise ValueError(
            "Normalizations are required in the configuration file (datamodule.normalizations)"
        )

    norm_configs = cfg["datamodule"]["normalizations"]
    normalizations = [load_normalization_from_config(norm) for norm in norm_configs]
    dm_params["normalizations"] = normalizations

    # Copy all other datamodule parameters
    for param, value in cfg["datamodule"].items():
        if param != "normalizations":
            # Handle patch sizes
            if param == "patch_size":
                dm_params["initial_yx_patch_size"] = value
                dm_params["final_yx_patch_size"] = value
            else:
                dm_params[param] = value

    # Set up the data module
    logger.info("Setting up data module")
    dm = TripletDataModule(**dm_params)

    # Get model parameters for handling 5D inputs
    channel_reduction_methods = {}

    if "model" in cfg and "channel_reduction_methods" in cfg["model"]:
        channel_reduction_methods = cfg["model"]["channel_reduction_methods"]

    # Initialize ImageNet model with reduction settings
    logger.info(f"Loading ImageNet model: {model}")
    model_module = ImageNetModule(
        model_name=model,
        channel_reduction_methods=channel_reduction_methods,
        channel_names=dm_params.get("source_channel", []),
    )

    # Get dimensionality reduction parameters from config
    phate_kwargs = None
    pca_kwargs = None

    if "embedding" in cfg:
        # Check for both capitalization variants and normalize
        if "phate_kwargs" in cfg["embedding"]:
            phate_kwargs = cfg["embedding"]["phate_kwargs"]

        if "umap_kwargs" in cfg["embedding"]:
            umap_kwargs = cfg["embedding"]["umap_kwargs"]

        if "pca_kwargs" in cfg["embedding"]:
            pca_kwargs = cfg["embedding"]["pca_kwargs"]

    # Check if output path exists and should be overwritten
    if "output_path" not in cfg["paths"]:
        raise ValueError(
            "Output path is required in the configuration file (paths.output_path)"
        )

    output_path = Path(cfg["paths"]["output_path"])
    output_dir = output_path.parent
    output_dir.mkdir(parents=True, exist_ok=True)

    overwrite = False
    if "execution" in cfg and "overwrite" in cfg["execution"]:
        overwrite = cfg["execution"]["overwrite"]
    elif output_path.exists():
        logger.warning(f"Output path {output_path} already exists, will overwrite")
        overwrite = True

    # Set up EmbeddingWriter callback
    embedding_writer = EmbeddingWriter(
        output_path=output_path,
        phate_kwargs=phate_kwargs,
        pca_kwargs=pca_kwargs,
        overwrite=overwrite,
    )

    # Set up and run VisCy trainer
    logger.info("Setting up VisCy trainer")
    trainer = VisCyTrainer(
        accelerator="gpu" if torch.cuda.is_available() else "cpu",
        devices=1,
        callbacks=[embedding_writer],
        inference_mode=True,
    )

    logger.info(f"Running prediction and saving to {output_path}")
    trainer.predict(model_module, datamodule=dm)

    # Save configuration if requested
    save_config_flag = True
    show_config_flag = True

    if "execution" in cfg:
        if "save_config" in cfg["execution"]:
            save_config_flag = cfg["execution"]["save_config"]
        if "show_config" in cfg["execution"]:
            show_config_flag = cfg["execution"]["show_config"]

    # Save configuration if requested
    if save_config_flag:
        config_path = os.path.join(output_dir, "config.yml")
        with open(config_path, "w") as f:
            yaml.dump(cfg, f, default_flow_style=False)
        logger.info(f"Configuration saved to {config_path}")

    # Display configuration if requested
    if show_config_flag:
        click.echo("\nConfiguration used:")
        click.echo("-" * 40)
        for key, value in cfg.items():
            click.echo(f"{key}:")
            if isinstance(value, dict):
                for subkey, subvalue in value.items():
                    if isinstance(subvalue, list) and subkey == "normalizations":
                        click.echo(f"  {subkey}:")
                        for norm in subvalue:
                            click.echo(f"    - class_path: {norm['class_path']}")
                            click.echo(f"      init_args: {norm['init_args']}")
                    else:
                        click.echo(f"  {subkey}: {subvalue}")
            else:
                click.echo(f"  {value}")
        click.echo("-" * 40)

    logger.info("Done!")


if __name__ == "__main__":
    main()



import logging
import math
import os
import re
import tempfile
from pathlib import Path
from typing import Callable, Literal, Sequence

import numpy as np
import torch
import zarr
from imageio import imread
from iohub.ngff import ImageArray, Plate, Position, open_ome_zarr
from lightning.pytorch import LightningDataModule
from monai.data import set_track_meta
from monai.data.utils import collate_meta_tensor
from monai.transforms import (
    CenterSpatialCropd,
    Compose,
    MapTransform,
    MultiSampleTrait,
    RandAffined,
)
from torch import Tensor
from torch.utils.data import DataLoader, Dataset

from viscy.data.typing import ChannelMap, DictTransform, HCSStackIndex, NormMeta, Sample

_logger = logging.getLogger("lightning.pytorch")


def _ensure_channel_list(str_or_seq: str | Sequence[str]) -> list[str]:
    """
    Ensure channel argument is a list of strings.

    :param Union[str, Sequence[str]] str_or_seq: channel name or list of channel names
    :return list[str]: list of channel names
    """
    if isinstance(str_or_seq, str):
        return [str_or_seq]
    try:
        return list(str_or_seq)
    except TypeError:
        raise TypeError(
            "Channel argument must be a string or sequence of strings. "
            f"Got {str_or_seq}."
        )


def _search_int_in_str(pattern: str, file_name: str) -> str:
    """Search image indices in a file name with regex patterns and strip leading zeros.
    E.g. ``'001'`` -> ``1``"""
    match = re.search(pattern, file_name)
    if match:
        return match.group()
    else:
        raise ValueError(f"Cannot find pattern {pattern} in {file_name}.")


def _collate_samples(batch: Sequence[Sample]) -> Sample:
    """Collate samples into a batch sample.

    :param Sequence[Sample] batch: a sequence of dictionaries,
        where each key may point to a value of a single tensor or a list of tensors,
        as is the case with ``train_patches_per_stack > 1``.
    :return Sample: Batch sample (dictionary of tensors)
    """
    collated: Sample = {}
    for key in batch[0].keys():
        data = []
        for sample in batch:
            if isinstance(sample[key], Sequence):
                data.extend(sample[key])
            else:
                data.append(sample[key])
        collated[key] = collate_meta_tensor(data)
    return collated


def _read_norm_meta(fov: Position) -> NormMeta | None:
    """
    Read normalization metadata from the FOV.
    Convert to float32 tensors to avoid automatic casting to float64.
    """
    norm_meta = fov.zattrs.get("normalization", None)
    if norm_meta is None:
        return None
    for channel, channel_values in norm_meta.items():
        for level, level_values in channel_values.items():
            for stat, value in level_values.items():
                if isinstance(value, Tensor):
                    value = value.clone().float()
                else:
                    value = torch.tensor(value, dtype=torch.float32)
                norm_meta[channel][level][stat] = value
    return norm_meta


class SlidingWindowDataset(Dataset):
    """Torch dataset where each element is a window of
    (C, Z, Y, X) where C=2 (source and target) and Z is ``z_window_size``.

    :param list[Position] positions: FOVs to include in dataset
    :param ChannelMap channels: source and target channel names,
        e.g. ``{'source': 'Phase', 'target': ['Nuclei', 'Membrane']}``
    :param int z_window_size: Z window size of the 2.5D U-Net, 1 for 2D
    :param DictTransform | None transform:
        a callable that transforms data, defaults to None
    """

    def __init__(
        self,
        positions: list[Position],
        channels: ChannelMap,
        z_window_size: int,
        transform: DictTransform | None = None,
    ) -> None:
        super().__init__()
        self.positions = positions
        self.channels = {k: _ensure_channel_list(v) for k, v in channels.items()}
        self.source_ch_idx = [
            positions[0].get_channel_index(c) for c in channels["source"]
        ]
        self.target_ch_idx = (
            [positions[0].get_channel_index(c) for c in channels["target"]]
            if "target" in channels
            else None
        )
        self.z_window_size = z_window_size
        self.transform = transform
        self._get_windows()

    def _get_windows(self) -> None:
        """Count the sliding windows along T and Z,
        and build an index-to-window LUT."""
        w = 0
        self.window_keys = []
        self.window_arrays = []
        self.window_norm_meta: list[NormMeta | None] = []
        for fov in self.positions:
            img_arr: ImageArray = fov["0"]
            ts = img_arr.frames
            zs = img_arr.slices - self.z_window_size + 1
            if zs < 1:
                raise IndexError(
                    f"Z window size {self.z_window_size} "
                    f"is larger than the number of Z slices ({img_arr.slices}) "
                    f"for FOV {fov.name}."
                )
            w += ts * zs
            self.window_keys.append(w)
            self.window_arrays.append(img_arr)
            self.window_norm_meta.append(_read_norm_meta(fov))
        self._max_window = w

    def _find_window(self, index: int) -> tuple[ImageArray, int, NormMeta | None]:
        """Look up window given index."""
        window_idx = sorted(self.window_keys + [index + 1]).index(index + 1)
        w = self.window_keys[window_idx]
        tz = index - self.window_keys[window_idx - 1] if window_idx > 0 else index
        norm_meta = self.window_norm_meta[self.window_keys.index(w)]
        return (self.window_arrays[self.window_keys.index(w)], tz, norm_meta)

    def _read_img_window(
        self, img: ImageArray, ch_idx: list[int], tz: int
    ) -> tuple[list[Tensor], HCSStackIndex]:
        """Read image window as tensor.

        :param ImageArray img: NGFF image array
        :param list[int] ch_idx: list of channel indices to read,
            output channel ordering will reflect the sequence
        :param int tz: window index within the FOV, counted Z-first
        :return list[Tensor], HCSStackIndex:
            list of (C=1, Z, Y, X) image tensors,
            tuple of image name, time index, and Z index
        """
        zs = img.shape[-3] - self.z_window_size + 1
        t = (tz + zs) // zs - 1
        z = tz - t * zs
        data = img.oindex[
            slice(t, t + 1),
            [int(i) for i in ch_idx],
            slice(z, z + self.z_window_size),
        ].astype(np.float32)
        return torch.from_numpy(data).unbind(dim=1), (img.name, t, z)

    def __len__(self) -> int:
        return self._max_window

    # TODO: refactor to a top level function
    def _stack_channels(
        self,
        sample_images: list[dict[str, Tensor]] | dict[str, Tensor],
        key: str,
    ) -> Tensor | list[Tensor]:
        """Stack single-channel images into a multi-channel tensor."""
        if not isinstance(sample_images, list):
            return torch.stack([sample_images[ch][0] for ch in self.channels[key]])
        # training time
        return [
            torch.stack([im[ch][0] for ch in self.channels[key]])
            for im in sample_images
        ]

    def __getitem__(self, index: int) -> Sample:
        img, tz, norm_meta = self._find_window(index)
        ch_names = self.channels["source"].copy()
        ch_idx = self.source_ch_idx.copy()
        if self.target_ch_idx is not None:
            ch_names.extend(self.channels["target"])
            ch_idx.extend(self.target_ch_idx)
        images, sample_index = self._read_img_window(img, ch_idx, tz)
        sample_images = {k: v for k, v in zip(ch_names, images)}
        if self.target_ch_idx is not None:
            # FIXME: this uses the first target channel as weight for performance
            # since adding a reference to a tensor does not copy
            # maybe write a weight map in preprocessing to use more information?
            sample_images["weight"] = sample_images[self.channels["target"][0]]
        if norm_meta is not None:
            sample_images["norm_meta"] = norm_meta
        if self.transform:
            sample_images = self.transform(sample_images)
        if "weight" in sample_images:
            del sample_images["weight"]
        sample = {
            "index": sample_index,
            "source": self._stack_channels(sample_images, "source"),
            "norm_meta": norm_meta,
        }
        if self.target_ch_idx is not None:
            sample["target"] = self._stack_channels(sample_images, "target")
        return sample


class MaskTestDataset(SlidingWindowDataset):
    """Torch dataset where each element is a window of
    (C, Z, Y, X) where C=2 (source and target) and Z is ``z_window_size``.
    This a testing stage version of :py:class:`viscy.data.hcs.SlidingWindowDataset`,
    and can only be used with batch size 1 for efficiency (no padding for collation),
    since the mask is not available for each stack.

    :param list[Position] positions: FOVs to include in dataset
    :param ChannelMap channels: source and target channel names,
        e.g. ``{'source': 'Phase', 'target': ['Nuclei', 'Membrane']}``
    :param int z_window_size: Z window size of the 2.5D U-Net, 1 for 2D
    :param DictTransform transform:
        a callable that transforms data, defaults to None
    :param str | None ground_truth_masks: path to the ground truth masks
    """

    def __init__(
        self,
        positions: list[Position],
        channels: ChannelMap,
        z_window_size: int,
        transform: DictTransform | None = None,
        ground_truth_masks: str | None = None,
    ) -> None:
        super().__init__(positions, channels, z_window_size, transform)
        self.masks = {}
        for img_path in Path(ground_truth_masks).glob("*cp_masks.png"):
            img_name = img_path.name
            position_name = _search_int_in_str(r"(?<=_p)\d{3}", img_name)
            # TODO: specify time index in the file name
            t_idx = 0
            # TODO: record channel name
            # channel_name = re.search(r"^.+(?=_p\d{3})", img_name).group()
            z_idx = _search_int_in_str(r"(?<=_z)\d+", img_name)
            self.masks[(int(position_name), int(t_idx), int(z_idx))] = img_path
        _logger.info(str(self.masks))

    def __getitem__(self, index: int) -> Sample:
        sample = super().__getitem__(index)
        img_name, t_idx, z_idx = sample["index"]
        position_name = int(img_name.split("/")[-2])
        key = (position_name, int(t_idx), int(z_idx) + self.z_window_size // 2)
        if img_path := self.masks.get(key):
            sample["labels"] = torch.from_numpy(imread(img_path).astype(np.int16))
        return sample


class HCSDataModule(LightningDataModule):
    """
    Lightning data module for a preprocessed HCS NGFF Store.

    Parameters
    ----------
    data_path : str
        Path to the data store.
    source_channel : str or Sequence[str]
        Name(s) of the source channel, e.g. 'Phase'.
    target_channel : str or Sequence[str]
        Name(s) of the target channel, e.g. ['Nuclei', 'Membrane'].
    z_window_size : int
        Z window size of the 2.5D U-Net, 1 for 2D.
    split_ratio : float, optional
        Split ratio of the training subset in the fit stage,
        e.g. 0.8 means an 80/20 split between training/validation,
        by default 0.8.
    batch_size : int, optional
        Batch size, defaults to 16.
    num_workers : int, optional
        Number of data-loading workers, defaults to 8.
    target_2d : bool, optional
        Whether the target is 2D (e.g. in a 2.5D model),
        defaults to False.
    yx_patch_size : tuple[int, int], optional
        Patch size in (Y, X), defaults to (256, 256).
    normalizations : list of MapTransform, optional
        MONAI dictionary transforms applied to selected channels,
        defaults to ``[]`` (no normalization).
    augmentations : list of MapTransform, optional
        MONAI dictionary transforms applied to the training set,
        defaults to ``[]`` (no augmentation).
    caching : bool, optional
        Whether to decompress all the images and cache the result,
        will store in `/tmp/$SLURM_JOB_ID/` if available,
        defaults to False.
    ground_truth_masks : Path or None, optional
        Path to the ground truth masks,
        used in the test stage to compute segmentation metrics,
        defaults to None.
    persistent_workers : bool, optional
        Whether to keep the workers alive between fitting epochs,
        defaults to False.
    prefetch_factor : int or None, optional
        Number of samples loaded in advance by each worker during fitting,
        defaults to None (2 per PyTorch default).
    """

    def __init__(
        self,
        data_path: str,
        source_channel: str | Sequence[str],
        target_channel: str | Sequence[str],
        z_window_size: int,
        split_ratio: float = 0.8,
        batch_size: int = 16,
        num_workers: int = 8,
        target_2d: bool = False,
        yx_patch_size: tuple[int, int] = (256, 256),
        normalizations: list[MapTransform] = [],
        augmentations: list[MapTransform] = [],
        caching: bool = False,
        ground_truth_masks: Path | None = None,
        persistent_workers=False,
        prefetch_factor=None,
        pin_memory=False,
    ):
        super().__init__()
        self.data_path = Path(data_path)
        self.source_channel = _ensure_channel_list(source_channel)
        self.target_channel = _ensure_channel_list(target_channel)
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.target_2d = target_2d
        self.z_window_size = z_window_size
        self.split_ratio = split_ratio
        self.yx_patch_size = yx_patch_size
        self.normalizations = normalizations
        self.augmentations = augmentations
        self.caching = caching
        self.ground_truth_masks = ground_truth_masks
        self.prepare_data_per_node = True
        self.persistent_workers = persistent_workers
        self.prefetch_factor = prefetch_factor
        self.pin_memory = pin_memory

    @property
    def cache_path(self):
        return Path(
            tempfile.gettempdir(),
            os.getenv("SLURM_JOB_ID", "viscy_cache"),
            self.data_path.name,
        )

    @property
    def maybe_cached_data_path(self):
        return self.cache_path if self.caching else self.data_path

    def _data_log_path(self) -> Path:
        log_dir = Path.cwd()
        if self.trainer:
            if self.trainer.logger:
                if self.trainer.logger.log_dir:
                    log_dir = Path(self.trainer.logger.log_dir)
        log_dir.mkdir(parents=True, exist_ok=True)
        return log_dir / "data.log"

    def prepare_data(self):
        if not self.caching:
            return
        # setup logger
        logger = logging.getLogger("viscy_data")
        logger.propagate = False
        logger.setLevel(logging.DEBUG)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.INFO)
        logger.addHandler(console_handler)
        file_handler = logging.FileHandler(self._data_log_path())
        file_handler.setLevel(logging.DEBUG)
        logger.addHandler(file_handler)
        logger.info(f"Caching dataset at {self.cache_path}.")
        tmp_store = zarr.NestedDirectoryStore(self.cache_path)
        with open_ome_zarr(self.data_path, mode="r") as lazy_plate:
            _, skipped, _ = zarr.copy(
                lazy_plate.zgroup,
                zarr.open(tmp_store, mode="a"),
                name="/",
                log=logger.debug,
                if_exists="skip_initialized",
                compressor=None,
            )
        if skipped > 0:
            logger.warning(
                f"Skipped {skipped} items when caching. Check debug log for details."
            )

    @property
    def _base_dataset_settings(self) -> dict[str, dict[str, list[str]] | int]:
        return {
            "channels": {"source": self.source_channel},
            "z_window_size": self.z_window_size,
        }

    def setup(self, stage: Literal["fit", "validate", "test", "predict"]):
        dataset_settings = self._base_dataset_settings
        if stage in ("fit", "validate"):
            self._setup_fit(dataset_settings)
        elif stage == "test":
            self._setup_test(dataset_settings)
        elif stage == "predict":
            self._setup_predict(dataset_settings)
        else:
            raise NotImplementedError(f"{stage} stage")

    def _set_fit_global_state(self, num_positions: int) -> torch.Tensor:
        # disable metadata tracking in MONAI for performance
        set_track_meta(False)
        # shuffle positions, randomness is handled globally
        return torch.randperm(num_positions)

    def _setup_fit(self, dataset_settings: dict):
        """Set up the training and validation datasets."""
        train_transform, val_transform = self._fit_transform()
        dataset_settings["channels"]["target"] = self.target_channel
        data_path = self.maybe_cached_data_path
        plate = open_ome_zarr(data_path, mode="r")

        # shuffle positions, randomness is handled globally
        positions = [pos for _, pos in plate.positions()]
        shuffled_indices = self._set_fit_global_state(len(positions))
        positions = list(positions[i] for i in shuffled_indices)
        num_train_fovs = int(len(positions) * self.split_ratio)
        # training set needs to sample more Z range for augmentation
        train_dataset_settings = dataset_settings.copy()
        z_scale_low, z_scale_high = self.train_z_scale_range
        if z_scale_high <= 0.0:
            expanded_z = self.z_window_size
        else:
            expanded_z = math.ceil(self.z_window_size * (1 + z_scale_high))
            expanded_z -= expanded_z % 2
        train_dataset_settings["z_window_size"] = expanded_z
        # train/val split
        self.train_dataset = SlidingWindowDataset(
            positions[:num_train_fovs],
            transform=train_transform,
            **train_dataset_settings,
        )
        self.val_dataset = SlidingWindowDataset(
            positions[num_train_fovs:],
            transform=val_transform,
            **dataset_settings,
        )

    def _setup_test(self, dataset_settings: dict):
        """Set up the test stage."""
        if self.batch_size != 1:
            _logger.warning(f"Ignoring batch size {self.batch_size} in test stage.")

        dataset_settings["channels"]["target"] = self.target_channel
        data_path = self.maybe_cached_data_path
        plate = open_ome_zarr(data_path, mode="r")
        test_transform = Compose(self.normalizations)
        if self.ground_truth_masks:
            self.test_dataset = MaskTestDataset(
                [p for _, p in plate.positions()],
                transform=test_transform,
                ground_truth_masks=self.ground_truth_masks,
                **dataset_settings,
            )
        else:
            self.test_dataset = SlidingWindowDataset(
                [p for _, p in plate.positions()],
                transform=test_transform,
                **dataset_settings,
            )

    def _set_predict_global_state(self) -> None:
        # track metadata for inverting transform
        set_track_meta(True)
        if self.caching:
            _logger.warning("Ignoring caching config in 'predict' stage.")

    def _positions_maybe_single(self) -> list[Position]:
        dataset: Plate | Position = open_ome_zarr(self.data_path, mode="r")
        if isinstance(dataset, Position):
            try:
                plate_path = self.data_path.parent.parent.parent
                fov_name = self.data_path.relative_to(plate_path).as_posix()
                plate = open_ome_zarr(plate_path)
            except Exception:
                raise FileNotFoundError(
                    "Parent HCS store not found for single FOV input."
                )
            positions = [plate[fov_name]]
        elif isinstance(dataset, Plate):
            positions = [p for _, p in dataset.positions()]
        return positions

    def _setup_predict(
        self,
        dataset_settings: dict,
    ):
        """Set up the predict stage."""
        self._set_predict_global_state()
        predict_transform = Compose(self.normalizations)
        self.predict_dataset = SlidingWindowDataset(
            positions=self._positions_maybe_single(),
            transform=predict_transform,
            **dataset_settings,
        )

    def on_before_batch_transfer(self, batch: Sample, dataloader_idx: int) -> Sample:
        """Removes redundant Z slices if the target is 2D to save VRAM."""
        predicting = False
        if self.trainer:
            if self.trainer.predicting:
                predicting = True
        if predicting or isinstance(batch, Tensor):
            # skipping example input array
            return batch
        if self.target_2d:
            # slice the center during training or testing
            z_index = self.z_window_size // 2
            batch["target"] = batch["target"][:, :, slice(z_index, z_index + 1)]
        return batch

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size // self.train_patches_per_stack,
            num_workers=self.num_workers,
            shuffle=True,
            prefetch_factor=self.prefetch_factor if self.num_workers else None,
            persistent_workers=self.persistent_workers,
            collate_fn=_collate_samples,
            drop_last=True,
            pin_memory=self.pin_memory,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
            prefetch_factor=self.prefetch_factor if self.num_workers else None,
            persistent_workers=self.persistent_workers,
            pin_memory=self.pin_memory,
        )

    def test_dataloader(self):
        return DataLoader(
            self.test_dataset,
            batch_size=1,
            num_workers=self.num_workers,
            shuffle=False,
        )

    def predict_dataloader(self):
        return DataLoader(
            self.predict_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=False,
        )

    def _fit_transform(self) -> tuple[Compose, Compose]:
        """(normalization -> maybe augmentation -> center crop)
        Deterministic center crop as the last step of training and validation."""
        # TODO: These have a fixed order for now... ()
        final_crop = [
            CenterSpatialCropd(
                keys=self.source_channel + self.target_channel,
                roi_size=(
                    self.z_window_size,
                    self.yx_patch_size[0],
                    self.yx_patch_size[1],
                ),
            )
        ]
        train_transform = Compose(
            self.normalizations + self._train_transform() + final_crop
        )
        val_transform = Compose(self.normalizations + final_crop)
        return train_transform, val_transform

    def _train_transform(self) -> list[Callable]:
        """Setup training augmentations: check input values,
        and parse the number of Z slices and patches to sample per stack."""
        self.train_patches_per_stack = 1
        z_scale_range = None
        if self.augmentations:
            for aug in self.augmentations:
                if isinstance(aug, RandAffined):
                    if z_scale_range is not None:
                        raise ValueError(
                            "Only one RandAffined augmentation is allowed."
                        )
                    z_scale_range = aug.rand_affine.rand_affine_grid.scale_range[0]
                if isinstance(aug, MultiSampleTrait):
                    # e.g. RandWeightedCropd.cropper.num_samples
                    # this trait does not have any concrete interface
                    # so this attribute may not be the same for other transforms
                    num_samples = aug.cropper.num_samples
                    if self.batch_size % num_samples != 0:
                        raise ValueError(
                            "Batch size must be divisible by `num_samples` per stack. "
                            f"Got batch size {self.batch_size} and "
                            f"number of samples {num_samples} for "
                            f"transform type {type(aug)}."
                        )
                    self.train_patches_per_stack = num_samples
        else:
            self.augmentations = []
        if z_scale_range is not None:
            if isinstance(z_scale_range, (float, int)):
                z_scale_range = float(z_scale_range)
                z_scale_range = (-z_scale_range, z_scale_range)
            if z_scale_range[0] > 0 or z_scale_range[1] < 0:
                raise ValueError(f"Invalid scaling range: {z_scale_range}")
            self.train_z_scale_range = z_scale_range
        else:
            self.train_z_scale_range = (0.0, 0.0)
        _logger.debug(f"Training augmentations: {self.augmentations}")
        return list(self.augmentations)



from viscy.transforms._redef import (
    CenterSpatialCropd,
    Decollated,
    RandAdjustContrastd,
    RandAffined,
    RandFlipd,
    RandGaussianNoised,
    RandGaussianSmoothd,
    RandScaleIntensityd,
    RandSpatialCropd,
    RandWeightedCropd,
    ScaleIntensityRangePercentilesd,
    ToDeviced,
)
from viscy.transforms._transforms import (
    BatchedRandAffined,
    BatchedScaleIntensityRangePercentilesd,
    BatchedZoom,
    NormalizeSampled,
    RandGaussianNoiseTensord,
    RandInvertIntensityd,
    StackChannelsd,
    TiledSpatialCropSamplesd,
)

__all__ = [
    "BatchedRandAffined",
    "BatchedScaleIntensityRangePercentilesd",
    "BatchedZoom",
    "CenterSpatialCropd",
    "Decollated",
    "NormalizeSampled",
    "RandAdjustContrastd",
    "RandAffined",
    "RandFlipd",
    "RandGaussianNoised",
    "RandGaussianNoiseTensord",
    "RandGaussianSmoothd",
    "RandInvertIntensityd",
    "RandScaleIntensityd",
    "RandSpatialCropd",
    "RandWeightedCropd",
    "ScaleIntensityRangePercentilesd",
    "StackChannelsd",
    "TiledSpatialCropSamplesd",
    "ToDeviced",
]



import os
import sys

import iohub.ngff as ngff
import numpy as np
import pandas as pd

import viscy.utils.mp_utils as mp_utils
from viscy.utils.cli_utils import show_progress_bar


def write_meta_field(position: ngff.Position, metadata, field_name, subfield_name):
    """
    Writes 'metadata' to position's plate-level or FOV level .zattrs metadata by either
    creating a new field (field_name) according to 'metadata', or updating the metadata
    to an existing field if found,
    or concatenating the metadata from different channels.

    Assumes that the zarr store group given follows the OMG-NGFF HCS
    format as specified here:
            https://ngff.openmicroscopy.org/latest/#hcs-layout

    Warning: Dangerous. Writing metadata fields above the image-level of
            an HCS hierarchy can break HCS compatibility

    :param Position zarr_dir: NGFF position node object
    :param dict metadata: metadata dictionary to write to JSON .zattrs
    :param str subfield_name: name of subfield inside the the main field
        (values for different channels)
    """
    if field_name in position.zattrs:
        if subfield_name in position.zattrs[field_name]:
            position.zattrs[field_name][subfield_name].update(metadata)
        else:
            D1 = position.zattrs[field_name]
            field_metadata = {
                subfield_name: metadata,
            }
            # position.zattrs[field_name][subfield_name] = metadata
            position.zattrs[field_name] = {**D1, **field_metadata}
    else:
        field_metadata = {
            subfield_name: metadata,
        }
        position.zattrs[field_name] = field_metadata


def generate_normalization_metadata(
    zarr_dir,
    num_workers=4,
    channel_ids=-1,
    grid_spacing=32,
):
    """
    Generate pixel intensity metadata to be later used in on-the-fly normalization
    during training and inference. Sampling is used for efficient estimation of median
    and interquartile range for intensity values on both a dataset and field-of-view
    level.

    Normalization values are recorded in the image-level metadata in the corresponding
    position of each zarr_dir store. Format of metadata is as follows:
    {
        channel_idx : {
            dataset_statistics: dataset level normalization values (positive float),
            fov_statistics: field-of-view level normalization values (positive float)
        },
        .
        .
        .
    }

    :param str zarr_dir: path to zarr store directory containing dataset.
    :param int num_workers: number of cpu workers for multiprocessing, defaults to 4
    :param list/int channel_ids: indices of channels to process in dataset arrays,
                                    by default calculates all
    :param int grid_spacing: distance between points in sampling grid
    """
    plate = ngff.open_ome_zarr(zarr_dir, mode="r+")
    position_map = list(plate.positions())

    if channel_ids == -1:
        channel_ids = range(len(plate.channel_names))
    elif isinstance(channel_ids, int):
        channel_ids = [channel_ids]

    # get arguments for multiprocessed grid sampling
    mp_grid_sampler_args = []
    for _, position in position_map:
        mp_grid_sampler_args.append([position, grid_spacing])

    # sample values and use them to get normalization statistics
    for i, channel in enumerate(channel_ids):
        show_progress_bar(
            dataloader=channel_ids,
            current=i,
            process="sampling channel values",
        )

        channel_name = plate.channel_names[channel]
        this_channels_args = tuple([args + [channel] for args in mp_grid_sampler_args])

        # NOTE: Doing sequential mp with pool execution creates synchronization
        #      points between each step. This could be detrimental to performance
        positions, fov_sample_values = mp_utils.mp_sample_im_pixels(
            this_channels_args, num_workers
        )
        dataset_sample_values = np.concatenate(
            [arr.flatten() for arr in fov_sample_values]
        )
        fov_level_statistics = mp_utils.mp_get_val_stats(fov_sample_values, num_workers)
        dataset_level_statistics = mp_utils.get_val_stats(dataset_sample_values)

        dataset_statistics = {
            "dataset_statistics": dataset_level_statistics,
        }

        write_meta_field(
            position=plate,
            metadata=dataset_statistics,
            field_name="normalization",
            subfield_name=channel_name,
        )

        for j, pos in enumerate(positions):
            show_progress_bar(
                dataloader=position_map,
                current=j,
                process=f"calculating channel statistics {channel}/{list(channel_ids)}",
            )
            position_statistics = dataset_statistics | {
                "fov_statistics": fov_level_statistics[j],
            }

            write_meta_field(
                position=pos,
                metadata=position_statistics,
                field_name="normalization",
                subfield_name=channel_name,
            )
    plate.close()


def compute_zscore_params(
    frames_meta, ints_meta, input_dir, normalize_im, min_fraction=0.99
):
    """
    Get zscore median and interquartile range

    :param pd.DataFrame frames_meta: Dataframe containing all metadata
    :param pd.DataFrame ints_meta: Metadata containing intensity statistics
        each z-slice and foreground fraction for masks
    :param str input_dir: Directory containing images
    :param None or str normalize_im: normalization scheme for input images
    :param float min_fraction: Minimum foreground fraction (in case of masks)
        for computing intensity statistics.

    :return pd.DataFrame frames_meta: Dataframe containing all metadata
    :return pd.DataFrame ints_meta: Metadata containing intensity statistics
        each z-slice
    """

    assert normalize_im in [
        None,
        "slice",
        "volume",
        "dataset",
    ], 'normalize_im must be None or "slice" or "volume" or "dataset"'

    if normalize_im is None:
        # No normalization
        frames_meta["zscore_median"] = 0
        frames_meta["zscore_iqr"] = 1
        return frames_meta
    elif normalize_im == "dataset":
        agg_cols = ["time_idx", "channel_idx", "dir_name"]
    elif normalize_im == "volume":
        agg_cols = ["time_idx", "channel_idx", "dir_name", "pos_idx"]
    else:
        agg_cols = ["time_idx", "channel_idx", "dir_name", "pos_idx", "slice_idx"]
    # median and inter-quartile range are more robust than mean and std
    ints_meta_sub = ints_meta[ints_meta["fg_frac"] >= min_fraction]
    ints_agg_median = ints_meta_sub[agg_cols + ["intensity"]].groupby(agg_cols).median()
    ints_agg_hq = (
        ints_meta_sub[agg_cols + ["intensity"]].groupby(agg_cols).quantile(0.75)
    )
    ints_agg_lq = (
        ints_meta_sub[agg_cols + ["intensity"]].groupby(agg_cols).quantile(0.25)
    )
    ints_agg = ints_agg_median
    ints_agg.columns = ["zscore_median"]
    ints_agg["zscore_iqr"] = ints_agg_hq["intensity"] - ints_agg_lq["intensity"]
    ints_agg.reset_index(inplace=True)

    cols_to_merge = frames_meta.columns[
        [col not in ["zscore_median", "zscore_iqr"] for col in frames_meta.columns]
    ]
    frames_meta = pd.merge(
        frames_meta[cols_to_merge],
        ints_agg,
        how="left",
        on=agg_cols,
    )
    if frames_meta["zscore_median"].isnull().values.any():
        raise ValueError(
            "Found NaN in normalization parameters. \
        min_fraction might be too low or images might be corrupted."
        )
    frames_meta_filename = os.path.join(input_dir, "frames_meta.csv")
    frames_meta.to_csv(frames_meta_filename, sep=",")

    cols_to_merge = ints_meta.columns[
        [col not in ["zscore_median", "zscore_iqr"] for col in ints_meta.columns]
    ]
    ints_meta = pd.merge(
        ints_meta[cols_to_merge],
        ints_agg,
        how="left",
        on=agg_cols,
    )
    ints_meta["intensity_norm"] = (
        ints_meta["intensity"] - ints_meta["zscore_median"]
    ) / (ints_meta["zscore_iqr"] + sys.float_info.epsilon)

    return frames_meta, ints_meta



# %%
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np

from viscy.representation.embedding_writer import read_embedding_dataset
from viscy.representation.evaluation.distance import (
   calculate_normalized_euclidean_distance_cell,
   compute_displacement_mean_std_full,
)

# %% paths 

features_path_30_min = Path(
   "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)

feature_path_no_track = Path("/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr")

features_path_any_time = Path("/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_difcell_randomtime_sampling/Ver2_updateTracking_refineModel/predictions/Feb_2chan_128patch_32projDim/2chan_128patch_56ckpt_FebTest.zarr")

data_path = Path(
   "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/registered_test.zarr"
)

tracks_path = Path(
   "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/track_test.zarr"
)

# %% Load embedding datasets for all three sampling
fov_name = '/B/4/6'
track_id = 52

embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)
embedding_dataset_any_time = read_embedding_dataset(features_path_any_time)

#%%
# Calculate displacement for each sampling
time_points_30_min, cosine_similarities_30_min = calculate_normalized_euclidean_distance_cell(embedding_dataset_30_min, fov_name, track_id)
time_points_no_track, cosine_similarities_no_track = calculate_normalized_euclidean_distance_cell(embedding_dataset_no_track, fov_name, track_id)
time_points_any_time, cosine_similarities_any_time = calculate_normalized_euclidean_distance_cell(embedding_dataset_any_time, fov_name, track_id)

# %% Plot displacement over time for all three conditions

plt.figure(figsize=(10, 6))

plt.plot(time_points_no_track, cosine_similarities_no_track, marker='o', label='classical contrastive (no tracking)')
plt.plot(time_points_any_time, cosine_similarities_any_time, marker='o', label='cell aware')
plt.plot(time_points_30_min, cosine_similarities_30_min, marker='o', label='cell & time aware (interval 30 min)')

plt.xlabel("Time Delay (t)", fontsize=10)
plt.ylabel("Normalized Euclidean Distance with First Time Point", fontsize=10)
plt.title("Normalized Euclidean Distance (Features) Over Time for Infected Cell", fontsize=12)

plt.grid(True)
plt.legend(fontsize=10)

#plt.savefig('4_euc_dist_full.svg', format='svg')
plt.show()


# %% Paths to datasets
features_path_30_min = Path("/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr")
feature_path_no_track = Path("/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr")

embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)


# %%
max_tau = 10  

mean_displacement_30_min_euc, std_displacement_30_min_euc = compute_displacement_mean_std_full(embedding_dataset_30_min, max_tau)
mean_displacement_no_track_euc, std_displacement_no_track_euc = compute_displacement_mean_std_full(embedding_dataset_no_track, max_tau)

# %% Plot 2: Cosine Displacements
plt.figure(figsize=(10, 6))

taus = list(mean_displacement_30_min_euc.keys())

mean_values_30_min_euc = list(mean_displacement_30_min_euc.values())
std_values_30_min_euc = list(std_displacement_30_min_euc.values())

plt.plot(taus, mean_values_30_min_euc, marker='o', label='Cell & Time Aware (30 min interval)', color='green')
plt.fill_between(taus, 
                 np.array(mean_values_30_min_euc) - np.array(std_values_30_min_euc), 
                 np.array(mean_values_30_min_euc) + np.array(std_values_30_min_euc),
                 color='green', alpha=0.3, label='Std Dev (30 min interval)')

mean_values_no_track_euc = list(mean_displacement_no_track_euc.values())
std_values_no_track_euc = list(std_displacement_no_track_euc.values())

plt.plot(taus, mean_values_no_track_euc, marker='o', label='Classical Contrastive (No Tracking)', color='blue')
plt.fill_between(taus, 
                 np.array(mean_values_no_track_euc) - np.array(std_values_no_track_euc), 
                 np.array(mean_values_no_track_euc) + np.array(std_values_no_track_euc),
                 color='blue', alpha=0.3, label='Std Dev (No Tracking)')

plt.xlabel('Time Shift (τ)')
plt.ylabel('Euclidean Distance')
plt.title('Embedding Displacement Over Time (Features)')

plt.grid(True)
plt.legend()

plt.show()



"""Auxiliary utility functions"""

import iohub.ngff as ngff
import yaml


def _assert_unique_subset(subset, superset, name):
    """
    Helper function to allow for clean code:
        Throws error if unique elements of subset are not a subset of
        unique elements of superset.

    Returns unique elements of subset if given a list. If subset is -1,
    returns all unique elements of superset
    """
    if subset == -1:
        subset = superset
    if not (isinstance(subset, list) or isinstance(subset, tuple)):
        subset = list(subset)
    unique_subset = set(subset)
    unique_superset = set(superset)
    assert unique_subset.issubset(unique_superset), (
        f"{name} in requested {name}: {unique_subset}"
        f" not in available {name}: {unique_superset}"
    )
    return unique_subset


def validate_metadata_indices(
    zarr_dir,
    time_ids=[],
    channel_ids=[],
    slice_ids=[],
    pos_ids=[],
):
    """
    Check the availability of indices provided timepoints, channels, positions
    and slices for all data, and returns only the available of the specified
    indices.

    If input ids are None, the indices for that parameter will not be
    evaluated. If input ids are -1, all indices for that parameter will
    be returned.

    Assumes uniform structure, as such structure is required for HCS compatibility

    :param str zarr_dir: HCS-compatible zarr directory to validate indices against
    :param list time_ids: check availability of these timepoints in image
                                metadata
    :param list channel_ids: check availability of these channels in image
                                    metadata
    :param list pos_ids: Check availability of positions in zarr_dir
    :param list slice_ids: Check availability of z slices in image metadata

    :return dict indices_metadata: All indices found given input
    :raise AssertionError: If not all channels, timepoints, positions
        or slices are present
    """
    plate = ngff.open_ome_zarr(zarr_dir, layout="hcs", mode="r")
    position_path, position = next(plate.positions())

    # read available channel indices from zarr store
    available_time_ids = range(position.data.shape[0])
    if isinstance(channel_ids, int):
        available_channel_ids = range(len(plate.channel_names))
    elif isinstance(channel_ids[0], int):
        available_channel_ids = range(len(plate.channel_names))
    else:
        available_channel_ids = len(plate.channel_names)
    available_slice_ids = range(position.data.shape[-3])
    available_pos_ids = [x[0] for x in list(plate.positions())]

    # enforce that requested indices are subsets of available indices
    time_ids = _assert_unique_subset(time_ids, available_time_ids, "slices")
    channel_ids = _assert_unique_subset(channel_ids, available_channel_ids, "channels")
    slice_ids = _assert_unique_subset(slice_ids, available_slice_ids, "slices")
    pos_ids = _assert_unique_subset(pos_ids, available_pos_ids, "positions")

    indices_metadata = {
        "time_ids": list(time_ids),
        "channel_ids": list(channel_ids),
        "slice_ids": list(slice_ids),
        "pos_ids": list(pos_ids),
    }
    plate.close()
    return indices_metadata


def read_config(config_fname):
    """Read the config file in yml format

    :param str config_fname: fname of config yaml with its full path
    :return: dict config: Configuration parameters
    """

    with open(config_fname, "r") as f:
        config = yaml.safe_load(f)

    return config



import gradio as gr
import torch
from viscy.light.engine import VSUNet
from huggingface_hub import hf_hub_download
from numpy.typing import ArrayLike
import numpy as np
from skimage import exposure


class VSGradio:
    def __init__(self, model_config, model_ckpt_path):
        self.model_config = model_config
        self.model_ckpt_path = model_ckpt_path
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = None
        self.load_model()

    def load_model(self):
        # Load the model checkpoint and move it to the correct device (GPU or CPU)
        self.model = VSUNet.load_from_checkpoint(
            self.model_ckpt_path,
            architecture="UNeXt2_2D",
            model_config=self.model_config,
        )
        self.model.to(self.device)  # Move the model to the correct device (GPU/CPU)
        self.model.eval()

    def normalize_fov(self, input: ArrayLike):
        "Normalizing the fov with zero mean and unit variance"
        mean = np.mean(input)
        std = np.std(input)
        return (input - mean) / std

    def predict(self, inp):
        # Normalize the input and convert to tensor
        inp = self.normalize_fov(inp)
        inp = torch.from_numpy(np.array(inp).astype(np.float32))

        # Prepare the input dictionary and move input to the correct device (GPU or CPU)
        test_dict = dict(
            index=None,
            source=inp.unsqueeze(0).unsqueeze(0).unsqueeze(0).to(self.device),
        )

        # Run model inference
        with torch.inference_mode():
            self.model.on_predict_start()  # Necessary preprocessing for the model
            pred = (
                self.model.predict_step(test_dict, 0, 0).cpu().numpy()
            )  # Move output back to CPU for post-processing

        # Post-process the model output and rescale intensity
        nuc_pred = pred[0, 0, 0]
        mem_pred = pred[0, 1, 0]
        nuc_pred = exposure.rescale_intensity(nuc_pred, out_range=(0, 1))
        mem_pred = exposure.rescale_intensity(mem_pred, out_range=(0, 1))

        return nuc_pred, mem_pred


# Load the custom CSS from the file
def load_css(file_path):
    with open(file_path, "r") as file:
        return file.read()


# %%
if __name__ == "__main__":
    # Download the model checkpoint from Hugging Face
    model_ckpt_path = hf_hub_download(
        repo_id="compmicro-czb/VSCyto2D", filename="epoch=399-step=23200.ckpt"
    )

    # Model configuration
    model_config = {
        "in_channels": 1,
        "out_channels": 2,
        "encoder_blocks": [3, 3, 9, 3],
        "dims": [96, 192, 384, 768],
        "decoder_conv_blocks": 2,
        "stem_kernel_size": [1, 2, 2],
        "in_stack_depth": 1,
        "pretraining": False,
    }

    # Initialize the Gradio app using Blocks
    with gr.Blocks(css=load_css("style.css")) as demo:
        # Title and description
        gr.HTML(
            "<div class='title-block'>Image Translation (Virtual Staining) of cellular landmark organelles</div>"
        )
        # Improved description block with better formatting
        gr.HTML(
            """
            <div class='description-block'>
                <p><b>Model:</b> VSCyto2D</p>
                <p>
                    <b>Input:</b> label-free image (e.g., QPI or phase contrast) <br>
                    <b>Output:</b> two virtually stained channels: one for the <b>nucleus</b> and one for the <b>cell membrane</b>.
                </p>
                <p>
                    Check out our preprint: 
                    <a href='https://www.biorxiv.org/content/10.1101/2024.05.31.596901' target='_blank'><i>Liu et al.,Robust virtual staining of landmark organelles</i></a>
                </p>
            </div>
            """
        )

        vsgradio = VSGradio(model_config, model_ckpt_path)

        # Layout for input and output images
        with gr.Row():
            input_image = gr.Image(type="numpy", image_mode="L", label="Upload Image")
            with gr.Column():
                output_nucleus = gr.Image(type="numpy", label="VS Nucleus")
                output_membrane = gr.Image(type="numpy", label="VS Membrane")

        # Button to trigger prediction
        submit_button = gr.Button("Submit")

        # Define what happens when the button is clicked
        submit_button.click(
            vsgradio.predict,
            inputs=input_image,
            outputs=[output_nucleus, output_membrane],
        )

        # Example images and article
        gr.Examples(
            examples=["examples/a549.png", "examples/hek.png"], inputs=input_image
        )

        # Article or footer information
        gr.HTML(
            """
            <div class='article-block'>
            <p> Model trained primarily on HEK293T, BJ5, and A549 cells. For best results, use quantitative phase images (QPI) or Zernike phase contrast.</p>
            <p> For training, inference and evaluation of the model refer to the <a href='https://github.com/mehta-lab/VisCy/tree/main/examples/virtual_staining/dlmbl_exercise' target='_blank'>GitHub repository</a>.</p>
            </div>
            """
        )

    # Launch the Gradio app
    demo.launch()



import os
import unittest

import cv2
import numpy as np
import pandas as pd
from testfixtures import TempDirectory

import viscy.preprocessing.resize_images as resize_images
import viscy.utils.aux_utils as aux_utils


class TestResizeImages(unittest.TestCase):
    def setUp(self):
        """
        Set up a directory with some images to resample
        """
        self.tempdir = TempDirectory()
        self.temp_path = self.tempdir.path
        self.output_dir = os.path.join(self.temp_path, "out_dir")
        # Start frames meta file
        self.meta_name = "frames_meta.csv"
        self.frames_meta = aux_utils.make_dataframe()
        # Write images
        self.time_idx = 5
        self.slice_idx = 6
        self.pos_idx = 7
        self.im = 1500 * np.ones((30, 20), dtype=np.uint16)

        for c in range(4):
            for p in range(self.pos_idx, self.pos_idx + 2):
                im_name = aux_utils.get_im_name(
                    channel_idx=c,
                    slice_idx=self.slice_idx,
                    time_idx=self.time_idx,
                    pos_idx=p,
                )
                cv2.imwrite(os.path.join(self.temp_path, im_name), self.im + c * 100)
                self.frames_meta = self.frames_meta.append(
                    aux_utils.parse_idx_from_name(
                        im_name=im_name, dir_name=self.temp_path
                    ),
                    ignore_index=True,
                )
        # Write metadata
        self.frames_meta.to_csv(
            os.path.join(self.temp_path, self.meta_name),
            sep=",",
        )

    def tearDown(self):
        """
        Tear down temporary folder and file structure
        """
        TempDirectory.cleanup_all()
        nose.tools.assert_equal(os.path.isdir(self.temp_path), False)

    def test_downsample(self):
        # Half the image size
        scale_factor = 0.5
        resize_inst = resize_images.ImageResizer(
            input_dir=self.temp_path,
            output_dir=self.output_dir,
            scale_factor=scale_factor,
        )
        self.assertEqual(resize_inst.time_ids, self.time_idx)
        self.assertListEqual(resize_inst.channel_ids.tolist(), [0, 1, 2, 3])
        self.assertEqual(resize_inst.slice_ids, self.slice_idx)
        self.assertListEqual(resize_inst.pos_ids.tolist(), [7, 8])
        resize_dir = resize_inst.get_resize_dir()
        self.assertEqual(os.path.join(self.output_dir, "resized_images"), resize_dir)
        # Resize
        resize_inst.resize_frames()
        # Validate
        new_shape = tuple([int(scale_factor * x) for x in self.im.shape])
        for i, row in self.frames_meta.iterrows():
            file_name = os.path.join(resize_dir, row["file_name"])
            im = cv2.imread(file_name, cv2.IMREAD_ANYDEPTH)
            self.assertTupleEqual(new_shape, im.shape)
            self.assertEqual(im.dtype, self.im.dtype)
            im_expected = self.im + row["channel_idx"] * 100
            im_expected = cv2.resize(im_expected, (new_shape[1], new_shape[0]))
            np.testing.assert_array_equal(im, im_expected)

    def test_upsample(self):
        # Half the image size
        scale_factor = 2.0
        resize_inst = resize_images.ImageResizer(
            input_dir=self.temp_path,
            output_dir=self.output_dir,
            scale_factor=scale_factor,
        )
        self.assertEqual(resize_inst.time_ids, self.time_idx)
        self.assertListEqual(resize_inst.channel_ids.tolist(), [0, 1, 2, 3])
        self.assertEqual(resize_inst.slice_ids, self.slice_idx)
        self.assertListEqual(resize_inst.pos_ids.tolist(), [7, 8])
        resize_dir = resize_inst.get_resize_dir()
        self.assertEqual(os.path.join(self.output_dir, "resized_images"), resize_dir)
        # Resize
        resize_inst.resize_frames()
        # Validate
        new_shape = tuple([int(scale_factor * x) for x in self.im.shape])
        for i, row in self.frames_meta.iterrows():
            file_name = os.path.join(resize_dir, row["file_name"])
            im = cv2.imread(file_name, cv2.IMREAD_ANYDEPTH)
            self.assertTupleEqual(new_shape, im.shape)
            self.assertEqual(im.dtype, self.im.dtype)
            im_expected = self.im + row["channel_idx"] * 100
            im_expected = cv2.resize(im_expected, (new_shape[1], new_shape[0]))
            np.testing.assert_array_equal(im, im_expected)

    def test_resize_volumes(self):
        """Test resizing volumes"""

        # set up a volume with 5 slices, 2 channels
        slice_ids = [0, 1, 2, 3, 4]
        channel_ids = [2, 3]
        resize_dir = os.path.join(self.output_dir, "resized_images")
        frames_meta = aux_utils.make_dataframe()
        exp_meta_dict = []
        for c in channel_ids:
            for s in slice_ids:
                im_name = aux_utils.get_im_name(
                    channel_idx=c,
                    slice_idx=s,
                    time_idx=self.time_idx,
                    pos_idx=self.pos_idx,
                )
                cv2.imwrite(os.path.join(self.temp_path, im_name), self.im + c * 100)
                frames_meta = frames_meta.append(
                    aux_utils.parse_idx_from_name(
                        im_name=im_name, dir_name=self.temp_path
                    ),
                    ignore_index=True,
                )
            op_fname = "im_c00{}_z000_t005_p007_3.3-0.8-1.0.npy".format(c)
            exp_meta_dict.append(
                {
                    "time_idx": self.time_idx,
                    "pos_idx": self.pos_idx,
                    "channel_idx": c,
                    "slice_idx": 0,
                    "file_name": op_fname,
                    "mean": np.mean(self.im) + c * 100,
                    "std": float(0),
                    "dir_name": resize_dir,
                }
            )
        exp_meta_df = pd.DataFrame.from_dict(exp_meta_dict)
        # Write metadata
        frames_meta.to_csv(
            os.path.join(self.temp_path, self.meta_name),
            sep=",",
        )

        scale_factor = [3.3, 0.8, 1.0]
        resize_inst = resize_images.ImageResizer(
            input_dir=self.temp_path,
            output_dir=self.output_dir,
            scale_factor=scale_factor,
        )

        # save all slices in one volume
        resize_inst.resize_volumes()
        saved_meta = aux_utils.read_meta(resize_dir)
        pd.testing.assert_frame_equal(saved_meta, exp_meta_df)

        # num_slices_subvolume = 3, save vol chunks
        exp_meta_dict = []
        for c in channel_ids:
            for s in [0, 2]:
                op_fname = "im_c00{}_z00{}_t005_p007_3.3-0.8-1.0.npy".format(c, s)
                exp_meta_dict.append(
                    {
                        "time_idx": self.time_idx,
                        "pos_idx": self.pos_idx,
                        "channel_idx": c,
                        "slice_idx": s,
                        "file_name": op_fname,
                        "mean": np.mean(self.im) + c * 100,
                        "std": float(0),
                        "dir_name": resize_dir,
                    }
                )
        exp_meta_df = pd.DataFrame.from_dict(exp_meta_dict)
        resize_inst.resize_volumes(num_slices_subvolume=3)
        saved_meta = aux_utils.read_meta(resize_dir)
        pd.testing.assert_frame_equal(saved_meta, exp_meta_df)



# script to profile dataloading
# use with a sampling profiler like py-spy
from monai.transforms import (
    Decollated,
    RandAdjustContrastd,
    RandGaussianSmoothd,
    RandScaleIntensityd,
    ToDeviced,
)
from pytorch_metric_learning.losses import NTXentLoss

from viscy.data.combined import BatchedConcatDataModule
from viscy.data.triplet import TripletDataModule
from viscy.representation.engine import ContrastiveEncoder, ContrastiveModule
from viscy.transforms import (
    NormalizeSampled,
)
from viscy.transforms._transforms import (
    BatchedRandAffined,
    BatchedScaleIntensityRangePercentilesd,
    RandGaussianNoiseTensord,
)


def model(
    input_channel_number: int = 1,
    z_stack_depth: int = 30,
    patch_size: int = 192,
    temperature: float = 0.5,
):
    return ContrastiveModule(
        encoder=ContrastiveEncoder(
            backbone="convnext_tiny",
            in_channels=input_channel_number,
            in_stack_depth=z_stack_depth,
            stem_kernel_size=(5, 4, 4),
            embedding_dim=768,
            projection_dim=32,
            drop_path_rate=0.0,
        ),
        loss_function=NTXentLoss(temperature=temperature),
        lr=0.00002,
        log_batches_per_epoch=3,
        log_samples_per_batch=3,
        example_input_array_shape=[
            1,
            input_channel_number,
            z_stack_depth,
            patch_size,
            patch_size,
        ],
    )


def channel_augmentations(processing_channel: str):
    return [
        BatchedRandAffined(
            keys=[processing_channel],
            prob=0.8,
            scale_range=((1.0, 1.0), (0.8, 1.2), (0.8, 1.2)),
            rotate_range=[1.0, 0.0, 0.0],
            shear_range=(0.2, 0.2, 0.0, 0.2, 0.0, 0.2),
        ),
        Decollated(keys=[processing_channel]),
        RandAdjustContrastd(
            keys=[processing_channel],
            prob=0.5,
            gamma=[0.8, 1.2],
        ),
        RandScaleIntensityd(
            keys=[processing_channel],
            prob=0.5,
            factors=0.5,
        ),
        RandGaussianSmoothd(
            keys=[processing_channel],
            prob=0.5,
            sigma_x=[0.25, 0.75],
            sigma_y=[0.25, 0.75],
            sigma_z=[0.0, 0.0],
        ),
        RandGaussianNoiseTensord(
            keys=[processing_channel],
            prob=0.5,
            mean=0.0,
            std=0.2,
        ),
    ]


def channel_normalization(
    phase_channel: str = None,
    fl_channel: str = None,
):
    if phase_channel:
        return [
            NormalizeSampled(
                keys=[phase_channel],
                level="fov_statistics",
                subtrahend="mean",
                divisor="std",
            )
        ]
    elif fl_channel:
        return [
            ToDeviced(keys=[fl_channel], device="cuda"),
            BatchedScaleIntensityRangePercentilesd(
                keys=[fl_channel],
                lower=50,
                upper=99,
                b_min=0.0,
                b_max=1.0,
            ),
        ]
    else:
        raise NotImplementedError("Either phase_channel or fl_channel must be provided")


if __name__ == "__main__":
    dm1 = TripletDataModule(
        data_path="/hpc/projects/organelle_phenotyping/datasets/organelle/SEC61B/2024_10_16_A549_SEC61_ZIKV_DENV/2024_10_16_A549_SEC61_ZIKV_DENV_2.zarr",
        tracks_path="/hpc/projects/intracellular_dashboard/organelle_dynamics/rerun/2024_10_16_A549_SEC61_ZIKV_DENV/1-preprocess/label-free/3-track/2024_10_16_A549_SEC61_ZIKV_DENV_cropped.zarr",
        source_channel=["raw GFP EX488 EM525-45"],
        z_range=[5, 35],
        initial_yx_patch_size=(384, 384),
        final_yx_patch_size=(192, 192),
        batch_size=16,
        num_workers=4,
        time_interval=1,
        augmentations=channel_augmentations("raw GFP EX488 EM525-45"),
        normalizations=channel_normalization(
            phase_channel=None, fl_channel="raw GFP EX488 EM525-45"
        ),
        fit_include_wells=["B/3", "B/4", "C/3", "C/4"],
        return_negative=False,
    )
    dm2 = TripletDataModule(
        data_path="/hpc/projects/organelle_phenotyping/datasets/organelle/SEC61B/2024_10_16_A549_SEC61_ZIKV_DENV/2024_10_16_A549_SEC61_ZIKV_DENV_2.zarr",
        tracks_path="/hpc/projects/intracellular_dashboard/organelle_dynamics/rerun/2024_10_16_A549_SEC61_ZIKV_DENV/1-preprocess/label-free/3-track/2024_10_16_A549_SEC61_ZIKV_DENV_cropped.zarr",
        source_channel=["raw mCherry EX561 EM600-37"],
        z_range=[5, 35],
        initial_yx_patch_size=(384, 384),
        final_yx_patch_size=(192, 192),
        batch_size=16,
        num_workers=4,
        time_interval=1,
        augmentations=channel_augmentations("raw mCherry EX561 EM600-37"),
        normalizations=channel_normalization(
            phase_channel=None, fl_channel="raw mCherry EX561 EM600-37"
        ),
        fit_include_wells=["B/3", "B/4", "C/3", "C/4"],
        return_negative=False,
    )
    dm = BatchedConcatDataModule(data_modules=[dm1, dm2])
    dm.setup("fit")

    print(len(dm1.train_dataset), len(dm2.train_dataset), len(dm.train_dataset))
    n = 1

    print("Training batches:")
    for i, batch in enumerate(dm.train_dataloader()):
        print(i, batch["anchor"].shape, batch["positive"].device)
        if i == n - 1:
            break
    print("Validation batches:")
    for i, batch in enumerate(dm.val_dataloader()):
        print(i, batch["anchor"].shape, batch["positive"].device)
        if i == n - 1:
            break



# %%
# Import necessary libraries, try euclidean distance for both features and
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from sklearn.preprocessing import StandardScaler
from umap import UMAP

from viscy.representation.embedding_writer import read_embedding_dataset
from viscy.representation.evaluation.distance import (
    calculate_cosine_similarity_cell,
    compute_displacement,
    compute_displacement_mean_std,
)

# %% Paths and parameters.


features_path_30_min = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)


feature_path_no_track = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr"
)


features_path_any_time = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_difcell_randomtime_sampling/Ver2_updateTracking_refineModel/predictions/Feb_2chan_128patch_32projDim/2chan_128patch_56ckpt_FebTest.zarr"
)


data_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/registered_test.zarr"
)


tracks_path = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/2024_02_04_A549_DENV_ZIKV_timelapse/8-train-test-split/track_test.zarr"
)


# %% Load embedding datasets for all three sampling
fov_name = "/B/4/6"
track_id = 52

embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)
embedding_dataset_any_time = read_embedding_dataset(features_path_any_time)

# Calculate cosine similarities for each sampling
time_points_30_min, cosine_similarities_30_min = calculate_cosine_similarity_cell(
    embedding_dataset_30_min, fov_name, track_id
)
time_points_no_track, cosine_similarities_no_track = calculate_cosine_similarity_cell(
    embedding_dataset_no_track, fov_name, track_id
)
time_points_any_time, cosine_similarities_any_time = calculate_cosine_similarity_cell(
    embedding_dataset_any_time, fov_name, track_id
)

# %% Plot cosine similarities over time for all three conditions

plt.figure(figsize=(10, 6))

plt.plot(
    time_points_no_track,
    cosine_similarities_no_track,
    marker="o",
    label="classical contrastive (no tracking)",
)
plt.plot(
    time_points_any_time, cosine_similarities_any_time, marker="o", label="cell aware"
)
plt.plot(
    time_points_30_min,
    cosine_similarities_30_min,
    marker="o",
    label="cell & time aware (interval 30 min)",
)

plt.xlabel("Time Delay (t)")
plt.ylabel("Cosine Similarity with First Time Point")
plt.title("Cosine Similarity Over Time for Infected Cell")

# plt.savefig('infected_cell_example.pdf', format='pdf')


plt.grid(True)

plt.legend()

plt.savefig("new_example_cell.svg", format="svg")


plt.show()
# %%


# %% import statements


# %% Paths to datasets
features_path_30_min = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)
feature_path_no_track = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr"
)
# features_path_any_time = Path("/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_difcell_randomtime_sampling/Ver2_updateTracking_refineModel/predictions/Feb_1chan_128patch_32projDim/1chan_128patch_63ckpt_FebTest.zarr")


# %% Read embedding datasets
embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)
# embedding_dataset_any_time = read_embedding_dataset(features_path_any_time)


# %% Compute displacements for both datasets (using Euclidean distance and Cosine similarity)
max_tau = 10  # Maximum time shift (tau) to compute displacements


# mean_displacement_30_min, std_displacement_30_min = compute_displacement_mean_std(embedding_dataset_30_min, max_tau, use_cosine=False, use_dissimilarity=False)
# mean_displacement_no_track, std_displacement_no_track = compute_displacement_mean_std(embedding_dataset_no_track, max_tau, use_cosine=False, use_dissimilarity=False)
# mean_displacement_any_time, std_displacement_any_time = compute_displacement_mean_std(embedding_dataset_any_time, max_tau, use_cosine=False)


mean_displacement_30_min_cosine, std_displacement_30_min_cosine = (
    compute_displacement_mean_std(
        embedding_dataset_30_min, max_tau, use_cosine=True, use_dissimilarity=False
    )
)
mean_displacement_no_track_cosine, std_displacement_no_track_cosine = (
    compute_displacement_mean_std(
        embedding_dataset_no_track, max_tau, use_cosine=True, use_dissimilarity=False
    )
)
# mean_displacement_any_time_cosine, std_displacement_any_time_cosine = compute_displacement_mean_std(embedding_dataset_any_time, max_tau, use_cosine=True)
# %% Plot 1: Euclidean Displacements
plt.figure(figsize=(10, 6))


taus = list(mean_displacement_30_min_cosine.keys())
mean_values_30_min = list(mean_displacement_30_min_cosine.values())
std_values_30_min = list(std_displacement_30_min_cosine.values())


mean_values_no_track = list(mean_displacement_no_track_cosine.values())
std_values_no_track = list(std_displacement_no_track_cosine.values())


# mean_values_any_time = list(mean_displacement_any_time.values())
# std_values_any_time = list(std_displacement_any_time.values())


# Plotting Euclidean displacements
plt.plot(
    taus, mean_values_30_min, marker="o", label="Cell & Time Aware (30 min interval)"
)
plt.fill_between(
    taus,
    np.array(mean_values_30_min) - np.array(std_values_30_min),
    np.array(mean_values_30_min) + np.array(std_values_30_min),
    color="gray",
    alpha=0.3,
    label="Std Dev (30 min interval)",
)


plt.plot(
    taus, mean_values_no_track, marker="o", label="Classical Contrastive (No Tracking)"
)
plt.fill_between(
    taus,
    np.array(mean_values_no_track) - np.array(std_values_no_track),
    np.array(mean_values_no_track) + np.array(std_values_no_track),
    color="blue",
    alpha=0.3,
    label="Std Dev (No Tracking)",
)


plt.xlabel("Time Shift (τ)")
plt.ylabel("Displacement")
plt.title("Embedding Displacement Over Time")
plt.grid(True)
plt.legend()


# plt.savefig('embedding_displacement_euclidean.svg', format='svg')
# plt.savefig('embedding_displacement_euclidean.pdf', format='pdf')


# Show the Euclidean plot
plt.show()


# %% Plot 2: Cosine Displacements
plt.figure(figsize=(10, 6))

taus = list(mean_displacement_30_min_cosine.keys())

# Plotting Cosine displacements
mean_values_30_min_cosine = list(mean_displacement_30_min_cosine.values())
std_values_30_min_cosine = list(std_displacement_30_min_cosine.values())


mean_values_no_track_cosine = list(mean_displacement_no_track_cosine.values())
std_values_no_track_cosine = list(std_displacement_no_track_cosine.values())


plt.plot(
    taus,
    mean_values_30_min_cosine,
    marker="o",
    label="Cell & Time Aware (30 min interval)",
)
plt.fill_between(
    taus,
    np.array(mean_values_30_min_cosine) - np.array(std_values_30_min_cosine),
    np.array(mean_values_30_min_cosine) + np.array(std_values_30_min_cosine),
    color="gray",
    alpha=0.3,
    label="Std Dev (30 min interval)",
)


plt.plot(
    taus,
    mean_values_no_track_cosine,
    marker="o",
    label="Classical Contrastive (No Tracking)",
)
plt.fill_between(
    taus,
    np.array(mean_values_no_track_cosine) - np.array(std_values_no_track_cosine),
    np.array(mean_values_no_track_cosine) + np.array(std_values_no_track_cosine),
    color="blue",
    alpha=0.3,
    label="Std Dev (No Tracking)",
)


plt.xlabel("Time Shift (τ)")
plt.ylabel("Cosine Similarity")
plt.title("Embedding Displacement Over Time")


plt.grid(True)
plt.legend()
plt.savefig("1_std_cosine_plot.svg", format="svg")

# Show the Cosine plot
plt.show()
# %%


# %% Paths to datasets
features_path_30_min = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)
feature_path_no_track = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr"
)


# %% Read embedding datasets
embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)


# %% Compute displacements for both datasets (using Cosine similarity)
max_tau = 10  # Maximum time shift (tau) to compute displacements


# Compute displacements for Cell & Time Aware (30 min interval) using Cosine similarity
displacement_per_tau_aware_cosine = compute_displacement(
    embedding_dataset_30_min,
    max_tau,
    use_cosine=True,
    use_dissimilarity=False,
    use_umap=False,
)


# Compute displacements for Classical Contrastive (No Tracking) using Cosine similarity
displacement_per_tau_contrastive_cosine = compute_displacement(
    embedding_dataset_no_track,
    max_tau,
    use_cosine=True,
    use_dissimilarity=False,
    use_umap=False,
)


# %% Prepare data for violin plot
def prepare_violin_data(taus, displacement_aware, displacement_contrastive):
    # Create a list to hold the data
    data = []

    # Populate the data for Cell & Time Aware
    for tau in taus:
        displacements_aware = displacement_aware.get(tau, [])
        for displacement in displacements_aware:
            data.append(
                {
                    "Time Shift (τ)": tau,
                    "Displacement": displacement,
                    "Sampling": "Cell & Time Aware (30 min interval)",
                }
            )

    # Populate the data for Classical Contrastive
    for tau in taus:
        displacements_contrastive = displacement_contrastive.get(tau, [])
        for displacement in displacements_contrastive:
            data.append(
                {
                    "Time Shift (τ)": tau,
                    "Displacement": displacement,
                    "Sampling": "Classical Contrastive (No Tracking)",
                }
            )

    # Convert to a DataFrame
    df = pd.DataFrame(data)
    return df


taus = list(displacement_per_tau_aware_cosine.keys())


# Prepare the violin plot data
df = prepare_violin_data(
    taus, displacement_per_tau_aware_cosine, displacement_per_tau_contrastive_cosine
)


# Create a violin plot using seaborn
plt.figure(figsize=(12, 8))
sns.violinplot(
    x="Time Shift (τ)",
    y="Displacement",
    hue="Sampling",
    data=df,
    palette="Set2",
    scale="width",
    bw=0.2,
    inner=None,
    split=True,
    cut=0,
)


# Add labels and title
plt.xlabel("Time Shift (τ)", fontsize=14)
plt.ylabel("Cosine Similarity", fontsize=14)
plt.title("Cosine Similarity Distribution on Features", fontsize=16)
plt.grid(True, linestyle="--", alpha=0.6)  # Lighter grid lines for less distraction
plt.legend(title="Sampling", fontsize=12, title_fontsize=14)


# plt.ylim(0.5, 1.0)


# Save the violin plot as SVG and PDF
plt.savefig("1fixed_violin_plot_cosine_similarity.svg", format="svg")
# plt.savefig('violin_plot_cosine_similarity.pdf', format='pdf')


# Show the plot
plt.show()
# %% using umap violin plot

# %% Paths to datasets
features_path_30_min = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/time_interval/predict/feb_test_time_interval_1_epoch_178.zarr"
)
feature_path_no_track = Path(
    "/hpc/projects/intracellular_dashboard/viral-sensor/infection_classification/models/time_sampling_strategies/negpair_random_sampling2/feb_fixed_test_predict.zarr"
)

# %% Read embedding datasets
embedding_dataset_30_min = read_embedding_dataset(features_path_30_min)
embedding_dataset_no_track = read_embedding_dataset(feature_path_no_track)


# %% Compute UMAP on features
def compute_umap(dataset):
    features = dataset["features"]
    scaled_features = StandardScaler().fit_transform(features.values)
    umap = UMAP(n_components=2)  # Reduce to 2 dimensions
    embedding = umap.fit_transform(scaled_features)

    # Add UMAP coordinates using xarray functionality
    umap_features = features.assign_coords(
        UMAP1=("sample", embedding[:, 0]), UMAP2=("sample", embedding[:, 1])
    )
    return umap_features


# Apply UMAP to both datasets
umap_features_30_min = compute_umap(embedding_dataset_30_min)
umap_features_no_track = compute_umap(embedding_dataset_no_track)

# %%
print(umap_features_30_min)
# %% Visualize UMAP embeddings
# # Visualize UMAP embeddings for the 30 min interval
# plt.figure(figsize=(8, 6))
# plt.scatter(umap_features_30_min[:, 0], umap_features_30_min[:, 1], c=embedding_dataset_30_min["t"].values, cmap='viridis')
# plt.colorbar(label='Timepoints')
# plt.title('UMAP Projection of Features (30 min Interval)')
# plt.xlabel('UMAP1')
# plt.ylabel('UMAP2')
# plt.show()

# # Visualize UMAP embeddings for the No Tracking dataset
# plt.figure(figsize=(8, 6))
# plt.scatter(umap_features_no_track[:, 0], umap_features_no_track[:, 1], c=embedding_dataset_no_track["t"].values, cmap='viridis')
# plt.colorbar(label='Timepoints')
# plt.title('UMAP Projection of Features (No Tracking)')
# plt.xlabel('UMAP1')
# plt.ylabel('UMAP2')
# plt.show()
# %% Compute displacements using UMAP coordinates (using Cosine similarity)
max_tau = 10  # Maximum time shift (tau) to compute displacements

# Compute displacements for UMAP-processed Cell & Time Aware (30 min interval)
displacement_per_tau_aware_umap_cosine = compute_displacement(
    umap_features_30_min,
    max_tau,
    use_cosine=True,
    use_dissimilarity=False,
    use_umap=True,
)

# Compute displacements for UMAP-processed Classical Contrastive (No Tracking)
displacement_per_tau_contrastive_umap_cosine = compute_displacement(
    umap_features_no_track,
    max_tau,
    use_cosine=True,
    use_dissimilarity=False,
    use_umap=True,
)


# %% Prepare data for violin plot
def prepare_violin_data(taus, displacement_aware, displacement_contrastive):
    # Create a list to hold the data
    data = []

    # Populate the data for Cell & Time Aware
    for tau in taus:
        displacements_aware = displacement_aware.get(tau, [])
        for displacement in displacements_aware:
            data.append(
                {
                    "Time Shift (τ)": tau,
                    "Displacement": displacement,
                    "Sampling": "Cell & Time Aware (30 min interval)",
                }
            )

    # Populate the data for Classical Contrastive
    for tau in taus:
        displacements_contrastive = displacement_contrastive.get(tau, [])
        for displacement in displacements_contrastive:
            data.append(
                {
                    "Time Shift (τ)": tau,
                    "Displacement": displacement,
                    "Sampling": "Classical Contrastive (No Tracking)",
                }
            )

    # Convert to a DataFrame
    df = pd.DataFrame(data)
    return df


taus = list(displacement_per_tau_aware_umap_cosine.keys())

# Prepare the violin plot data
df = prepare_violin_data(
    taus,
    displacement_per_tau_aware_umap_cosine,
    displacement_per_tau_contrastive_umap_cosine,
)

# %% Create a violin plot using seaborn
plt.figure(figsize=(12, 8))
sns.violinplot(
    x="Time Shift (τ)",
    y="Displacement",
    hue="Sampling",
    data=df,
    palette="Set2",
    scale="width",
    bw=0.2,
    inner=None,
    split=True,
    cut=0,
)

# Add labels and title
plt.xlabel("Time Shift (τ)", fontsize=14)
plt.ylabel("Cosine Similarity", fontsize=14)
plt.title("Cosine Similarity Distribution using UMAP Features", fontsize=16)
plt.grid(True, linestyle="--", alpha=0.6)  # Lighter grid lines for less distraction
plt.legend(title="Sampling", fontsize=12, title_fontsize=14)

# plt.ylim(0, 1)

# Save the violin plot as SVG and PDF
plt.savefig("fixed_plot_cosine_similarity.svg", format="svg")
# plt.savefig('violin_plot_cosine_similarity_umap.pdf', format='pdf')

# Show the plot
plt.show()


# %%
# %% Visualize Displacement Distributions (Example Code)
# Compare displacement distributions for τ = 1
# plt.figure(figsize=(10, 6))
# sns.histplot(displacement_per_tau_aware_umap_cosine[1], kde=True, label='UMAP - 30 min Interval', color='blue')
# sns.histplot(displacement_per_tau_contrastive_umap_cosine[1], kde=True, label='UMAP - No Tracking', color='green')
# plt.legend()
# plt.title('Comparison of Displacement Distributions for τ = 1 (UMAP)')
# plt.xlabel('Displacement')
# plt.show()

# # Compare displacement distributions for the full feature set (same τ = 1)
# plt.figure(figsize=(10, 6))
# sns.histplot(displacement_per_tau_aware_cosine[1], kde=True, label='Full Features - 30 min Interval', color='red')
# sns.histplot(displacement_per_tau_contrastive_cosine[1], kde=True, label='Full Features - No Tracking', color='orange')
# plt.legend()
# plt.title('Comparison of Displacement Distributions for τ = 1 (Full Features)')
# plt.xlabel('Displacement')
# plt.show()
# # %%



import torch

from viscy.unet.networks.fcmae import (
    FullyConvolutionalMAE,
    MaskedAdaptiveProjection,
    MaskedConvNeXtV2Block,
    MaskedConvNeXtV2Stage,
    MaskedMultiscaleEncoder,
    PixelToVoxelShuffleHead,
    generate_mask,
    masked_patchify,
    masked_unpatchify,
    upsample_mask,
)


def test_generate_mask():
    w = 64
    s = 16
    m = 0.75
    mask = generate_mask((2, 3, w, w), stride=s, mask_ratio=m, device="cpu")
    assert mask.shape == (2, 1, w // s, w // s)
    assert mask.dtype == torch.bool
    ratio = mask.sum((2, 3)) / mask.numel() * mask.shape[0]
    assert torch.allclose(ratio, torch.ones_like(ratio) * m)


def test_masked_patchify():
    b, c, h, w = 2, 3, 4, 8
    x = torch.rand(b, c, h, w)
    mask_ratio = 0.75
    mask = generate_mask(x.shape, stride=2, mask_ratio=mask_ratio, device=x.device)
    mask = upsample_mask(mask, x.shape)
    feat = masked_patchify(x, ~mask)
    assert feat.shape == (b, int(h * w * (1 - mask_ratio)), c)


def test_unmasked_patchify_roundtrip():
    x = torch.rand(2, 3, 4, 8)
    y = masked_unpatchify(masked_patchify(x), out_shape=x.shape)
    assert torch.allclose(x, y)


def test_masked_patchify_roundtrip():
    x = torch.rand(2, 3, 4, 8)
    mask = generate_mask(x.shape, stride=2, mask_ratio=0.5, device=x.device)
    mask = upsample_mask(mask, x.shape)
    y = masked_unpatchify(masked_patchify(x, ~mask), out_shape=x.shape, unmasked=~mask)
    assert torch.all((y == 0) ^ (x == y))
    assert torch.all((y == 0)[:, 0:1] == mask)


def test_masked_convnextv2_block() -> None:
    x = torch.rand(2, 3, 4, 5)
    mask = generate_mask(x.shape, stride=1, mask_ratio=0.5, device=x.device)
    block = MaskedConvNeXtV2Block(3, 3 * 2)
    unmasked_out = block(x)
    assert len(unmasked_out.unique()) == x.numel() * 2
    all_unmasked = torch.ones_like(mask)
    empty_masked_out = block(x, all_unmasked)
    assert torch.allclose(unmasked_out, empty_masked_out)
    block = MaskedConvNeXtV2Block(3, 3)
    masked_out = block(x, mask)
    assert len(masked_out.unique()) == mask.sum() * x.shape[1] + 1


def test_masked_convnextv2_stage():
    x = torch.rand(2, 3, 16, 16)
    mask = generate_mask(x.shape, stride=4, mask_ratio=0.5, device=x.device)
    stage = MaskedConvNeXtV2Stage(3, 3, kernel_size=7, stride=2, num_blocks=2)
    out = stage(x)
    assert out.shape == (2, 3, 8, 8)
    masked_out = stage(x, mask)
    assert not torch.allclose(masked_out, out)


def test_adaptive_projection():
    proj = MaskedAdaptiveProjection(
        3, 12, kernel_size_2d=4, kernel_depth=5, in_stack_depth=5
    )
    assert proj(torch.rand(2, 3, 5, 8, 8)).shape == (2, 12, 2, 2)
    assert proj(torch.rand(2, 3, 1, 12, 16)).shape == (2, 12, 3, 4)
    mask = generate_mask((1, 3, 5, 8, 8), stride=4, mask_ratio=0.6, device="cpu")
    masked_out = proj(torch.rand(1, 3, 5, 16, 16), mask)
    assert masked_out.shape == (1, 12, 4, 4)
    proj = MaskedAdaptiveProjection(
        3, 12, kernel_size_2d=(2, 4), kernel_depth=5, in_stack_depth=15
    )
    assert proj(torch.rand(2, 3, 15, 6, 8)).shape == (2, 12, 3, 2)


def test_masked_multiscale_encoder():
    xy_size = 64
    dims = [12, 24, 48, 96]
    x = torch.rand(2, 3, 5, xy_size, xy_size)
    encoder = MaskedMultiscaleEncoder(3, dims=dims)
    auto_masked_features, _ = encoder(x, mask_ratio=0.5)
    target_shape = list(x.shape)
    target_shape.pop(1)
    assert len(auto_masked_features) == 4
    for i, (dim, afeat) in enumerate(zip(dims, auto_masked_features)):
        assert afeat.shape[0] == x.shape[0]
        assert afeat.shape[1] == dim
        stride = 2 * 2 ** (i + 1)
        assert afeat.shape[2] == afeat.shape[3] == xy_size // stride


def test_pixel_to_voxel_shuffle_head():
    head = PixelToVoxelShuffleHead(240, 3, out_stack_depth=5, xy_scaling=4)
    x = torch.rand(2, 240, 16, 16)
    y = head(x)
    assert y.shape == (2, 3, 5, 64, 64)


def test_fcmae():
    x = torch.rand(2, 3, 5, 128, 128)
    model = FullyConvolutionalMAE(3, 3)
    y, m = model(x)
    assert y.shape == x.shape
    assert m is None
    y, m = model(x, mask_ratio=0.6)
    assert y.shape == x.shape
    assert m.shape == (2, 1, 128, 128)


def test_fcmae_head_conv():
    x = torch.rand(2, 3, 5, 128, 128)
    model = FullyConvolutionalMAE(
        3, 3, head_conv=True, head_conv_expansion_ratio=4, head_conv_pool=True
    )
    y, m = model(x)
    assert y.shape == x.shape
    assert m is None
    y, m = model(x, mask_ratio=0.6)
    assert y.shape == x.shape
    assert m.shape == (2, 1, 128, 128)



"""3D version of `kornia.augmentation._2d.intensity.gaussian_blur`."""

from typing import Any, Iterable

from kornia.augmentation import random_generator as rg
from kornia.augmentation._3d.intensity.base import IntensityAugmentationBase3D
from kornia.constants import BorderType
from kornia.filters import filter3d, get_gaussian_kernel3d
from monai.transforms import MapTransform, RandomizableTransform
from torch import Tensor


class RandomGaussianBlur(IntensityAugmentationBase3D):
    def __init__(
        self,
        kernel_size: tuple[int, int, int] | int,
        sigma: tuple[float, float, float] | Tensor,
        border_type: str = "reflect",
        same_on_batch: bool = False,
        p: float = 0.5,
        keepdim: bool = False,
    ) -> None:
        super().__init__(p=p, same_on_batch=same_on_batch, p_batch=1.0, keepdim=keepdim)

        self.flags = {
            "kernel_size": kernel_size,
            "border_type": BorderType.get(border_type),
        }
        self._param_generator = rg.RandomGaussianBlurGenerator(sigma)

    def apply_transform(
        self,
        input: Tensor,
        params: dict[str, Tensor],
        flags: dict[str, Any],
        transform: Tensor | None = None,
    ) -> Tensor:
        sigma = params["sigma"].unsqueeze(-1).expand(-1, 2)
        kernel = get_gaussian_kernel3d(
            kernel_size=self.flags["kernel_size"], sigma=sigma
        )
        return filter3d(input, kernel, border_type=self.flags["border_type"])


class BatchedRandGaussianBlurd(MapTransform, RandomizableTransform):
    def __init__(
        self,
        keys: str | Iterable[str],
        kernel_size: tuple[int, int] | int,
        sigma: tuple[float, float],
        border_type: str = "reflect",
        same_on_batch: bool = False,
        prob: float = 0.1,
        allow_missing_keys: bool = False,
    ) -> None:
        MapTransform.__init__(self, keys, allow_missing_keys=allow_missing_keys)
        RandomizableTransform.__init__(self, prob)
        self.filter = RandomGaussianBlur(
            kernel_size=kernel_size,
            sigma=sigma,
            border_type=border_type,
            same_on_batch=same_on_batch,
            p=prob,
        )

    def __call__(self, sample: dict[str, Tensor]) -> dict[str, Tensor]:
        self.randomize(None)
        if not self._do_transform:
            return sample
        for key in self.keys:
            if key in sample:
                sample[key] = -sample[key]
        return sample



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class ConvBlock3D(nn.Module):
    def __init__(
        self,
        in_filters,
        out_filters,
        dropout=False,
        norm="batch",
        residual=True,
        activation="relu",
        transpose=False,
        kernel_size=(3, 3, 3),
        num_repeats=3,
        filter_steps="first",
        layer_order="can",
        padding=None,
    ):
        """
        Convolutional block for lateral layers in Unet.
        This block only accepts tensors of dimensions in
        order [...,z,x,y] or [...,z,y,x]

        Format for layer initialization is as follows:
            if layer type specified
            => for number of layers
            => add layer to list of that layer type
        This is done to allow for dynamic layer number specification in the conv blocks,
        which allows us to change the parameter numbers of the network.

        Only 'same' convolutional padding is recommended,
        as the conv blocks are intended for deep Unets.
        However padding can be specified as follows:
        padding -> token{'same', 'valid', 'valid_stack'} or tuple(int) or int:
                    -> 'same': pads with same convolution
                    -> 'valid': pads for valid convolution on all dimensions
                    -> 'valid_stack': pads for valid convolution on xy dims (-1, -2),
                        same on z dim (-3).
                    -> tuple (int): pads above and below corresponding dimensions
                    -> int: pads above and below all dimensions

        :param int in_filters: number of images in in stack
        :param int out_filters: number of images in out stack
        :param float dropout: dropout probability (False => 0)
        :param str norm: normalization type: 'batch', 'instance'
        :param bool residual: as name
        :param str activation: activation function: 'relu', 'leakyrelu', 'elu', 'selu'
        :param bool transpose: as name
        :param int/tuple kernel_size: convolutional kernel size
        :param int num_repeats: as name
        :param str filter_steps: determines where in the block
            the filters inflate channels
            (learn abstraction information): 'linear','first','last'
        :param str layer_order: order of conv, norm, and act layers in block:
            'can', 'cna', etc.
            NOTE: for now conv must always come first as required by norm feature counts
        :paramn str/tuple(int)/tuple/None padding: convolutional padding,
            see docstring for details
        """

        super(ConvBlock3D, self).__init__()
        self.in_filters = in_filters
        self.out_filters = out_filters
        self.dropout = dropout
        self.norm = norm
        self.residual = residual
        self.activation = activation
        self.transpose = transpose
        self.num_repeats = num_repeats
        self.filter_steps = filter_steps
        self.layer_order = layer_order

        # ---- Handle Kernel ----#
        ks = kernel_size
        if isinstance(ks, int):
            assert ks % 2 == 1, "Kernel dims must be odd"
        elif isinstance(ks, tuple):
            for i in range(len(ks)):
                assert ks[i] % 2 == 1, "Kernel dims must be odd"
            assert i == 2, "kernel_size length must be 3"
        else:
            raise AttributeError("'kernel_size' must be either int or tuple")
        self.kernel_size = kernel_size

        # ---- Handle Padding ----#
        self.pad_type = "same"
        self.padding = (ks[2] // 2, ks[1] // 2, ks[0] // 2)
        if padding == "valid":
            self.padding = (0, 0, 0)
        elif self.padding == "valid_stack":  # note: deprecated
            ks = kernel_size
            self.padding = (0, 0, ks[0] // 2)
        elif isinstance(padding, tuple):
            self.padding = padding
        self.padding = tuple(self.padding[i // 2] for i in range(6)) + (0,) * 4

        # ----- Init Dropout -----#
        if self.dropout:
            self.drop_list = []
            for i in range(self.num_repeats):
                self.drop_list.append(nn.Dropout3d(self.dropout))
            self.register_modules(self.drop_list, "dropout")

        # ---- Init linear filter steps ----#
        steps = np.linspace(in_filters, out_filters, num_repeats + 1).astype(int)

        # ----- Init Normalization Layers -----#
        self.norm_list = [None for i in range(num_repeats)]
        if self.norm == "batch":
            for i in range(self.num_repeats):
                if self.filter_steps == "linear":
                    self.norm_list[i] = nn.BatchNorm3d(steps[i + 1])
                elif self.filter_steps == "first":
                    self.norm_list[i] = nn.BatchNorm3d(steps[-1])
                elif self.filter_steps == "last":
                    if i < self.num_repeats - 1:
                        self.norm_list[i] = nn.BatchNorm3d(steps[0])
                    else:
                        self.norm_list[i] = nn.BatchNorm3d(steps[-1])
        elif self.norm == "instance":
            for i in range(self.num_repeats):
                if self.filter_steps == "linear":
                    self.norm_list[i] = nn.InstanceNorm3d(steps[i + 1])
                elif self.filter_steps == "first":
                    self.norm_list[i] = nn.InstanceNorm3d(steps[-1])
                elif self.filter_steps == "last":
                    if i < self.num_repeats - 1:
                        self.norm_list[i] = nn.InstanceNorm3d(steps[0])
                    else:
                        self.norm_list[i] = nn.InstanceNorm3d(steps[-1])
        self.register_modules(self.norm_list, f"{norm}_norm")

        # ----- Init Conv Layers -----#
        #
        # init conv layers and determine transposition during convolution
        # The parameters governing the initiation logic flow are:
        #                 self.filter_steps
        #                 self.transpose
        #                 self.num_repeats
        # See above for definitions.
        # -------#

        self.conv_list = []
        if self.filter_steps == "linear":
            for i in range(self.num_repeats):
                depth_pair = (
                    (steps[i], steps[i + 1])
                    if i + 1 < num_repeats
                    else (steps[i], steps[-1])
                )
                if self.transpose:
                    self.conv_list.append(
                        nn.ConvTranspose3d(
                            depth_pair[0], depth_pair[1], kernel_size=kernel_size
                        )
                    )
                else:
                    self.conv_list.append(
                        nn.Conv3d(depth_pair[0], depth_pair[1], kernel_size=kernel_size)
                    )

        elif self.filter_steps == "first":
            if self.transpose:
                for i in range(self.num_repeats):
                    if i == 0:
                        self.conv_list.append(
                            nn.ConvTranspose3d(
                                in_filters, out_filters, kernel_size=kernel_size
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.ConvTranspose3d(
                                out_filters, out_filters, kernel_size=kernel_size
                            )
                        )
            else:
                for i in range(self.num_repeats):
                    if i == 0:
                        self.conv_list.append(
                            nn.Conv3d(in_filters, out_filters, kernel_size=kernel_size)
                        )
                    else:
                        self.conv_list.append(
                            nn.Conv3d(out_filters, out_filters, kernel_size=kernel_size)
                        )

        elif self.filter_steps == "last":
            if self.transpose:
                for i in range(self.num_repeats):
                    if i == self.num_repeats - 1:
                        self.conv_list.append(
                            nn.ConvTranspose3d(
                                in_filters, out_filters, kernel_size=kernel_size
                            )
                        )
                    else:
                        self.conv_list.append(
                            nn.ConvTranspose3d(
                                in_filters, in_filters, kernel_size=kernel_size
                            )
                        )
            else:
                for i in range(self.num_repeats):
                    if i == self.num_repeats - 1:
                        self.conv_list.append(
                            nn.Conv3d(in_filters, out_filters, kernel_size=kernel_size)
                        )
                    else:
                        self.conv_list.append(
                            nn.Conv3d(in_filters, in_filters, kernel_size=kernel_size)
                        )
        self.register_modules(self.conv_list, "Conv3d")

        # ----- Init Residual Layer -----#
        # Note that convolution is only used in residual layer
        # when block is shrinking feature space
        # Unregistered -- Not a learnable parameter
        self.resid_conv = nn.Conv3d(
            self.in_filters, self.out_filters, kernel_size=1, padding=0
        )

        # ----- Init Activation Layers -----#
        self.act_list = []
        if self.activation == "relu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.ReLU())
        elif self.activation == "leakyrelu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.LeakyReLU())
        elif self.activation == "elu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.ELU())
        elif self.activation == "selu":
            for i in range(self.num_repeats):
                self.act_list.append(nn.SELU())
        elif self.activation != "linear":
            raise NotImplementedError(
                f"Activation type {self.activation} not supported."
            )
        self.register_modules(self.act_list, f"{self.activation}_act")

    def forward(self, x):
        """
        Forward call of convolutional block

        Order of layers within the block is defined by the 'layer_order' parameter,
        which is a string of 'c's, 'a's and 'n's in reference to
        convolution, activation, and normalization layers.
        This sequence is repeated num_repeats times.

        Recommended layer order:   convolution -> activation -> normalization

        Regardless of layer order,
        the final layer sequence in the block always ends in activation.
        This allows for usage of passthrough layers
        or a final output activation function determined separately.

        Residual blocks:
            if input channels are greater than output channels,
            we use a 1x1 convolution on input to get desired feature channels
            if input channels are less than output channels,
            we zero-pad input channels to output channel size

        :param torch.tensor x: input tensor
        """
        x_0 = x
        for i in range(self.num_repeats):
            order = list(self.layer_order)
            while len(order) > 0:
                layer = order.pop(0)
                if layer == "c":
                    x = F.pad(x, self.padding, "constant", 0)
                    x = self.conv_list[i](x)
                    if self.dropout:
                        x = self.drop_list[i](x)
                elif layer == "a":
                    if i < self.num_repeats - 1 or self.activation != "linear":
                        x = self.act_list[i](x)
                elif layer == "n" and self.norm_list[i]:
                    x = self.norm_list[i](x)

        # residual summation comes after final activation/normalization
        if self.residual:
            # pad/collapse feature dimension
            if self.in_filters > self.out_filters:
                x_0 = self.resid_conv(x_0)
            elif self.in_filters < self.out_filters:
                x_0 = F.pad(
                    x_0,
                    (*[0] * 6, self.out_filters - self.in_filters, *[0] * 3),
                    mode="constant",
                    value=0,
                )

            # fit xy dimensions
            if self.pad_type == "valid_stack":
                lost = [dim // 2 * self.num_repeats for dim in self.kernel_size[1:]]
                x_0 = x_0[
                    ...,
                    lost[0] : x_0.shape[-2] - lost[0],
                    lost[1] : x_0.shape[-1] - lost[1],
                ]

            x = torch.add(x, x_0)

        return x

    def model(self):
        """
        Allows calling of parameters inside ConvBlock object:
        'ConvBlock.model().parameters()''

        Layer order:       convolution -> normalization -> activation

        We can make a list of layer modules and unpack them into nn.Sequential.
        Note: this is distinct from the forward call
            because we want to use the forward call with addition,
            since this is a residual block.
            The forward call performs the residual calculation,
            and all the parameters can be seen by the optimizer when given this model.
        """
        layers = []

        for i in range(self.num_repeats):
            layers.append(self.conv_list[i])
            if self.dropout:
                layers.append(self.drop_list[i])
            if self.norm[i]:
                layers.append(self.norm_list[i])
            if i < len(self.act_list):
                layers.append(self.act_list[i])

        return nn.Sequential(*layers)

    def register_modules(self, module_list, name):
        """
        Helper function that registers modules stored in a list to the model object
        so that the can be seen by PyTorch optimizer.

        Used to enable model graph creation
        with non-sequential model types and dynamic layer numbers

        :param list(torch.nn.module) module_list: list of modules to register
        :param str name: name of module type
        """
        for i, module in enumerate(module_list):
            self.add_module(f"{name}_{str(i)}", module)
